# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import datetime
import logging
import sys
import typing as t
import uuid
from collections import defaultdict
from functools import cached_property

import sqlglot
from sqlglot import Dialect, exp
from sqlglot.dialects.dialect import DialectType, NormalizationStrategy
from sqlglot.expressions import parse_identifier
from sqlglot.helper import ensure_list, seq_get
from sqlglot.optimizer import optimize
from sqlglot.optimizer.normalize_identifiers import normalize_identifiers
from sqlglot.optimizer.qualify import qualify as qualify_func
from sqlglot.optimizer.qualify_columns import (
    quote_identifiers as quote_identifiers_func,
)
from sqlglot.schema import MappingSchema

from sqlframe.base.catalog import _BaseCatalog
from sqlframe.base.dataframe import _BaseDataFrame
from sqlframe.base.normalize import normalize_dict
from sqlframe.base.readerwriter import _BaseDataFrameReader, _BaseDataFrameWriter
from sqlframe.base.udf import _BaseUDFRegistration
from sqlframe.base.util import (
    get_column_mapping_from_schema_input,
    normalize_string,
    verify_pandas_installed,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if t.TYPE_CHECKING:
    import pandas as pd
    from _typeshed.dbapi import DBAPIConnection, DBAPICursor

    from sqlframe.base._typing import ColumnLiterals, SchemaInput
    from sqlframe.base.column import Column
    from sqlframe.base.types import Row, StructType

    class DBAPIConnectionWithPandas(DBAPIConnection):
        def cursor(self) -> DBAPICursorWithPandas: ...

    class DBAPICursorWithPandas(DBAPICursor):
        def fetchdf(self) -> pd.DataFrame: ...

    CONN = t.TypeVar("CONN", bound=DBAPIConnectionWithPandas)
else:
    CONN = t.TypeVar("CONN")


logger = logging.getLogger(__name__)


CATALOG = t.TypeVar("CATALOG", bound=_BaseCatalog)
READER = t.TypeVar("READER", bound=_BaseDataFrameReader)
WRITER = t.TypeVar("WRITER", bound=_BaseDataFrameWriter)
DF = t.TypeVar("DF", bound=_BaseDataFrame)
UDF_REGISTRATION = t.TypeVar("UDF_REGISTRATION", bound=_BaseUDFRegistration)

_MISSING = "MISSING"


class _BaseSession(t.Generic[CATALOG, READER, WRITER, DF, CONN, UDF_REGISTRATION]):
    _instance = None
    _reader: t.Type[READER]
    _writer: t.Type[WRITER]
    _catalog: t.Type[CATALOG]
    _df: t.Type[DF]
    _udf_registration: t.Type[UDF_REGISTRATION]

    SANITIZE_COLUMN_NAMES = False

    def __init__(
        self,
        conn: t.Optional[CONN] = None,
        schema: t.Optional[MappingSchema] = None,
        *args,
        **kwargs,
    ):
        if not hasattr(self, "input_dialect"):
            self.input_dialect: Dialect = Dialect.get_or_raise(self.builder.DEFAULT_INPUT_DIALECT)
            self.output_dialect: Dialect = Dialect.get_or_raise(self.builder.DEFAULT_OUTPUT_DIALECT)
            self.execution_dialect: Dialect = Dialect.get_or_raise(
                self.builder.DEFAULT_EXECUTION_DIALECT
            )
            self.known_ids: t.Set[str] = set()
            self.known_branch_ids: t.Set[str] = set()
            self.known_sequence_ids: t.Set[str] = set()
            self.name_to_sequence_id_mapping: t.Dict[str, t.List[str]] = defaultdict(list)
            self.incrementing_id: int = 1
            self._last_loaded_file: t.Optional[str] = None
            self.temp_views: t.Dict[str, DF] = {}
        if not self._has_connection or conn:
            self._connection = conn
        if not getattr(self, "schema", None) or schema:
            self._schema = schema

    @property
    def read(self) -> READER:
        return self._reader(self)

    @cached_property
    def catalog(self) -> CATALOG:
        return self._catalog(self, self._schema)

    @property
    def _conn(self) -> CONN:
        if self._connection is None:
            raise ValueError("Connection not set")
        return self._connection

    @cached_property
    def _cur(self) -> DBAPICursorWithPandas:
        return self._conn.cursor()

    @property
    def default_time_format(self) -> str:
        return self.input_dialect.TIME_FORMAT.strip("'")

    def format_time(self, value: t.Optional[t.Union[Column, str]] = None) -> exp.Expression:
        from sqlframe.base.column import Column

        value = value or self.default_time_format
        if isinstance(value, Column):
            value = value.expression.this
        return self.input_dialect.format_time(f"'{value}'")  # type: ignore

    def format_execution_time(
        self, value: t.Optional[t.Union[Column, str]] = None
    ) -> exp.Expression:
        from sqlframe.base.column import Column

        if value is None:
            return exp.Literal.string(self.execution_dialect.TIME_FORMAT.strip("'"))

        if isinstance(value, Column):
            value = value.expression.this
        return exp.Literal.string(
            self.execution_dialect.generator()
            .format_time(
                exp.StrToTime(this=exp.Null(), format=self.input_dialect.format_time(f"'{value}'"))
            )
            .strip("'")  # type: ignore
        )

    def _sanitize_column_name(self, name: str) -> str:
        if self.SANITIZE_COLUMN_NAMES:
            return name.replace("(", "_").replace(")", "_")
        return name

    def table(self, tableName: str) -> DF:
        return self.read.table(tableName)

    def _create_df(self, *args, **kwargs) -> DF:
        return self._df(self, *args, **kwargs)

    def __new__(cls, *args, **kwargs):
        if _BaseSession._instance is None:
            _BaseSession._instance = super().__new__(cls)
        return _BaseSession._instance

    @property
    def _has_connection(self) -> bool:
        return hasattr(self, "_connection") and bool(self._connection)

    @property
    def udf(self) -> UDF_REGISTRATION:
        return self._udf_registration(self)

    def getActiveSession(self) -> Self:
        return self

    def range(self, *args):
        start = 0
        step = 1
        numPartitions = None
        if len(args) == 1:
            end = args[0]
        elif len(args) == 2:
            start, end = args
        elif len(args) == 3:
            start, end, step = args
        elif len(args) == 4:
            start, end, step, numPartitions = args
        else:
            raise ValueError(
                "range() takes 1 to 4 positional arguments but {} were given".format(len(args))
            )
        if numPartitions is not None:
            logger.warning("numPartitions is not supported")
        return self.createDataFrame([[x] for x in range(start, end, step)], schema={"id": "long"})

    def createDataFrame(
        self,
        data: t.Sequence[
            t.Union[
                t.Dict[str, ColumnLiterals],
                t.List[ColumnLiterals],
                t.Tuple[ColumnLiterals, ...],
                ColumnLiterals,
            ]
        ],
        schema: t.Optional[SchemaInput] = None,
        samplingRatio: t.Optional[float] = None,
        verifySchema: bool = False,
    ) -> DF:
        from sqlframe.base import functions as F
        from sqlframe.base.types import Row, StructType

        if samplingRatio is not None or verifySchema:
            raise NotImplementedError("Sampling Ratio and Verify Schema are not supported")
        if (
            schema is not None
            and not isinstance(schema, dict)
            and (
                not isinstance(schema, (StructType, str, list, tuple))
                or (isinstance(schema, (list, tuple)) and not isinstance(schema[0], str))
            )
        ):
            raise NotImplementedError("Only schema of either list or string of list supported")

        column_mapping: t.Mapping[str, t.Optional[exp.DataType]]
        if schema is not None:
            column_mapping = get_column_mapping_from_schema_input(
                schema, dialect=self.input_dialect
            )
        elif data:
            if isinstance(data[0], Row):
                column_mapping = {col_name.strip(): None for col_name in data[0].__fields__}
            elif isinstance(data[0], dict):
                column_mapping = {col_name.strip(): None for col_name in data[0]}
            else:
                column_mapping = {f"_{i}": None for i in range(1, len(data[0]) + 1)}  # type: ignore
        else:
            column_mapping = {}

        column_mapping = {
            normalize_identifiers(k, self.input_dialect).sql(dialect=self.input_dialect): v
            for k, v in column_mapping.items()
        }
        empty_df = not data
        rows = [[None] * len(column_mapping)] if empty_df else list(data)  # type: ignore

        def get_default_data_type(value: t.Any) -> t.Optional[str]:
            if isinstance(value, Row):
                row_types = []
                for row_name, row_dtype in zip(value.__fields__, value):
                    default_type = get_default_data_type(row_dtype)
                    if not default_type:
                        continue
                    row_types.append((row_name, default_type))
                return "struct<" + ", ".join(f"{k}: {v}" for (k, v) in row_types) + ">"
            elif isinstance(value, dict):
                sample_row = seq_get(list(value.items()), 0)
                if not sample_row:
                    return None
                key, value = sample_row
                default_key = get_default_data_type(key)
                default_value = get_default_data_type(value)
                if not default_key or not default_value:
                    return None
                return f"map<{default_key}, {default_value}>"
            elif isinstance(value, (list, set, tuple)):
                if not value:
                    return None
                default_type = get_default_data_type(next(iter(value)))
                if not default_type:
                    return None
                return f"array<{default_type}>"
            elif isinstance(value, bool):
                return "boolean"
            elif isinstance(value, bytes):
                return "binary"
            elif isinstance(value, int):
                return "bigint"
            elif isinstance(value, float):
                return "double"
            elif isinstance(value, datetime.datetime):
                if value.tzinfo:
                    return "timestamptz"
                return "timestamp"
            elif isinstance(value, datetime.date):
                return "date"
            elif isinstance(value, str):
                return "string"
            return None

        updated_mapping: t.Dict[str, t.Optional[exp.DataType]] = {}
        sample_row = rows[0]
        for i, (name, dtype) in enumerate(column_mapping.items()):
            if dtype is not None:
                updated_mapping[name] = dtype
                continue
            if isinstance(sample_row, Row):
                sample_row = sample_row.asDict()
            if isinstance(sample_row, dict):
                sample_row = normalize_dict(self, sample_row)
                default_data_type = get_default_data_type(sample_row[name])
                updated_mapping[name] = (
                    exp.DataType.build(default_data_type, dialect="spark")
                    if default_data_type
                    else None
                )
            else:
                default_data_type = get_default_data_type(sample_row[i])
                updated_mapping[name] = (
                    exp.DataType.build(default_data_type, dialect="spark")
                    if default_data_type
                    else None
                )
        column_mapping = updated_mapping
        data_expressions = []
        for row in rows:
            if isinstance(row, (list, tuple, dict)):
                if not row:
                    data_expressions.append(exp.tuple_(exp.Null()))
                    continue
                if isinstance(row, Row):
                    row = row.asDict()
                if isinstance(row, dict):
                    row = row.values()  # type: ignore
                data_expressions.append(exp.tuple_(*[F.lit(x).expression for x in row]))
            else:
                data_expressions.append(exp.tuple_(*[F.lit(row).expression]))

        if column_mapping:
            sel_columns = [
                (
                    F.col(name).cast(data_type).alias(name).expression
                    if data_type is not None
                    else F.col(name).expression
                )
                for name, data_type in column_mapping.items()
            ]
        else:
            sel_columns = [F.lit(None).expression]

        select_kwargs = {
            "expressions": sel_columns,
            "from": exp.From(
                this=exp.Values(
                    expressions=data_expressions,
                    alias=exp.TableAlias(
                        this=exp.to_identifier(self._auto_incrementing_name),
                        columns=[
                            exp.parse_identifier(col_name, dialect=self.input_dialect)
                            for col_name in column_mapping
                        ],
                    ),
                ),
            ),
        }

        sel_expression = exp.Select(**select_kwargs)
        if empty_df:
            sel_expression = sel_expression.where(exp.false())
        return self._create_df(sel_expression)

    def sql(
        self,
        sqlQuery: t.Union[str, exp.Expression],
        dialect: DialectType = None,
        qualify: bool = True,
    ) -> DF:
        dialect = Dialect.get_or_raise(dialect or self.input_dialect)
        expression = (
            sqlglot.parse_one(
                normalize_string(sqlQuery, from_dialect=dialect, is_query=True), read=dialect
            )
            if isinstance(sqlQuery, str)
            else sqlQuery
        )
        expression = sqlglot.parse_one(
            normalize_string(
                expression.sql(dialect=dialect),
                from_dialect=dialect,
                to_dialect=self.input_dialect,
                is_query=True,
            ),
            dialect=self.input_dialect,
        )
        if qualify:
            expression = qualify_func(
                expression,
                dialect=dialect,
                quote_identifiers=False,
                identify=False,
                schema=self.catalog._schema,
            )
        if self.temp_views:
            replacement_mapping = {}
            for table in expression.find_all(exp.Table):
                if not (df := self.temp_views.get(table.name)):
                    continue
                expression_ctes = {cte.alias_or_name: cte for cte in expression.ctes}  # type: ignore
                replacement_mapping[table] = df.expression.ctes[-1].alias_or_name
                ctes_to_add = []
                for cte in df.expression.ctes:
                    if cte.alias_or_name not in expression_ctes:
                        ctes_to_add.append(cte)
                expression.set("with", exp.With(expressions=expression.ctes + ctes_to_add))  # type: ignore

            def replace_temp_view_name_with_cte(node: exp.Expression) -> exp.Expression:
                if isinstance(node, exp.Table):
                    if node in replacement_mapping:
                        node.set("this", exp.to_identifier(replacement_mapping[node]))
                return node

            if replacement_mapping:
                expression = expression.transform(replace_temp_view_name_with_cte)

        if isinstance(expression, exp.Select):
            df = self._create_df(expression)
            df = df._convert_leaf_to_cte()
        elif isinstance(expression, (exp.Create, exp.Insert)):
            select_expression = expression.expression.copy()
            if isinstance(expression, exp.Insert):
                select_expression.set("with", expression.args.get("with"))
                expression.set("with", None)
            del expression.args["expression"]
            df = self._create_df(select_expression, output_expression_container=expression)  # type: ignore
            df = df._convert_leaf_to_cte()
        else:
            raise ValueError(
                "Unknown expression type provided in the SQL. Please create an issue with the SQL."
            )
        return df

    @property
    def _auto_incrementing_name(self) -> str:
        name = f"a{self.incrementing_id}"
        self.incrementing_id += 1
        return name

    @property
    def _random_branch_id(self) -> str:
        id = self._random_id
        self.known_branch_ids.add(id)
        return id

    @property
    def _random_sequence_id(self):
        id = self._random_id
        self.known_sequence_ids.add(id)
        return id

    @property
    def _random_id(self) -> str:
        id = "r" + uuid.uuid4().hex
        normalized_id = self._normalize_string(id)
        self.known_ids.add(normalized_id)
        return normalized_id

    @property
    def _join_hint_names(self) -> t.Set[str]:
        return {"BROADCAST", "MERGE", "SHUFFLE_HASH", "SHUFFLE_REPLICATE_NL"}

    def _normalize_string(self, value: str) -> str:
        expression = parse_identifier(value, dialect=self.input_dialect)
        normalize_identifiers(expression, dialect=self.input_dialect)
        return expression.sql(dialect=self.input_dialect)

    def _add_alias_to_mapping(self, name: str, sequence_id: str):
        self.name_to_sequence_id_mapping[self._normalize_string(name)].append(sequence_id)

    def _collect(
        self,
        expressions: t.Union[str, exp.Expression, t.List[str], t.List[exp.Expression]],
        *,
        quote_identifiers: bool = True,
        skip_normalization: bool = False,
        skip_rows: bool = False,
    ) -> t.List[Row]:
        for expression in ensure_list(expressions):
            if isinstance(expression, exp.Expression):
                sql = (
                    expression.sql(dialect=self.execution_dialect)
                    if skip_normalization
                    else self._to_sql(expression, quote_identifiers=quote_identifiers)
                )
            else:
                sql = expression  # type: ignore
            self._execute(sql)
        if skip_rows:
            return []
        result = self._cur.fetchall()
        if not self._cur.description:
            return []
        columns = [
            normalize_string(x[0], from_dialect="execution", to_dialect="output", is_column=True)
            for x in self._cur.description
        ]
        return [self._to_row(columns, row) for row in result]

    def _fetchdf(
        self,
        expressions: t.Union[exp.Expression, t.List[exp.Expression]],
        *,
        quote_identifiers: bool = True,
    ) -> pd.DataFrame:
        verify_pandas_installed()
        from pandas.io.sql import read_sql_query

        all_expressions = [None] + ensure_list(expressions)
        for an_expression in all_expressions[:-1]:
            if an_expression:
                self._execute(self._to_sql(an_expression, quote_identifiers=quote_identifiers))  # type: ignore
        assert all_expressions[-1] is not None
        return read_sql_query(
            self._to_sql(all_expressions[-1], quote_identifiers=quote_identifiers),  # type: ignore
            self._conn,
        )

    def _to_sql(
        self,
        sql: t.Union[str, exp.Expression],
        *,
        dialect: DialectType = None,
        quote_identifiers: bool = True,
        pretty: bool = False,
    ) -> str:
        return normalize_string(
            sql,
            from_dialect=self.input_dialect,
            to_dialect=dialect or self.execution_dialect,
            is_query=True,
            quote_identifiers=quote_identifiers,
            pretty=pretty,
        )

    def _optimize(
        self,
        expression: exp.Expression,
        dialect: t.Optional[Dialect] = None,
        quote_identifiers: bool = True,
    ) -> exp.Expression:
        dialect = dialect or self.input_dialect
        normalize_identifiers(expression, dialect=dialect)
        if quote_identifiers:
            quote_identifiers_func(expression, dialect=dialect)
        return optimize(expression, dialect=dialect, schema=self.catalog._schema)

    def _execute(self, sql: str) -> None:
        self._cur.execute(sql)

    @classmethod
    def _try_get_map(cls, value: t.Any) -> t.Optional[t.Dict[str, t.Any]]:
        return None if not isinstance(value, dict) else value

    @classmethod
    def _to_value(cls, value: t.Any) -> t.Any:
        if (map_value := cls._try_get_map(value)) is not None:
            return {
                normalize_string(k, from_dialect="execution", to_dialect="output")
                if isinstance(k, str)
                else k: cls._to_value(v)
                for k, v in map_value.items()
            }
        elif isinstance(value, dict):
            return cls._to_row(list(value.keys()), list(value.values()))
        elif isinstance(value, (list, set, tuple)) and value:
            return [cls._to_value(x) for x in value]
        elif isinstance(value, datetime.datetime):
            return value.replace(tzinfo=None)
        return value

    @classmethod
    def _to_row(cls, columns: t.List[str], values: t.Iterable[t.Any]) -> Row:
        from sqlframe.base.types import Row, _create_row

        converted_values = []
        for value in values:
            converted_values.append(cls._to_value(value))
        return _create_row(columns, converted_values)

    class Builder:
        SQLFRAME_INPUT_DIALECT_KEY = "sqlframe.input.dialect"
        SQLFRAME_OUTPUT_DIALECT_KEY = "sqlframe.output.dialect"
        SQLFRAME_EXECUTION_DIALECT_KEY = "sqlframe.execution.dialect"
        SQLFRAME_CONN_KEY = "sqlframe.conn"
        SQLFRAME_SCHEMA_KEY = "sqlframe.schema"
        DEFAULT_INPUT_DIALECT = "spark"
        DEFAULT_OUTPUT_DIALECT = "spark"
        DEFAULT_EXECUTION_DIALECT = "spark"

        def __init__(self):
            self.input_dialect = self.DEFAULT_INPUT_DIALECT
            self.output_dialect = self.DEFAULT_OUTPUT_DIALECT
            self.execution_dialect = self.DEFAULT_EXECUTION_DIALECT
            self._conn = None
            self._session_kwargs = {}

        def __getattr__(self, item) -> Self:
            return self

        def __call__(self, *args, **kwargs):
            return self

        @property
        def session(self) -> _BaseSession:
            return _BaseSession(**self._session_kwargs)

        def getOrCreate(self) -> _BaseSession:
            from sqlframe import ACTIVATE_CONFIG

            for k, v in ACTIVATE_CONFIG.items():
                self._set_config(k, v)
            self._set_session_properties()
            return self.session

        def _set_config(
            self,
            key: t.Optional[str] = None,
            value: t.Optional[t.Any] = None,
            *,
            map: t.Optional[t.Dict[str, t.Any]] = None,
        ) -> None:
            if value is not None:
                if key == self.SQLFRAME_INPUT_DIALECT_KEY:
                    self.input_dialect = value
                elif key == self.SQLFRAME_OUTPUT_DIALECT_KEY:
                    self.output_dialect = value
                elif key == self.SQLFRAME_EXECUTION_DIALECT_KEY:
                    self.execution_dialect = value
                elif key == self.SQLFRAME_CONN_KEY:
                    self._session_kwargs["conn"] = value
                elif key == self.SQLFRAME_SCHEMA_KEY:
                    self._session_kwargs["schema"] = value
                else:
                    self._session_kwargs[key] = value
            if map:
                if self.SQLFRAME_INPUT_DIALECT_KEY in map:
                    self.input_dialect = map[self.SQLFRAME_INPUT_DIALECT_KEY]
                if self.SQLFRAME_OUTPUT_DIALECT_KEY in map:
                    self.output_dialect = map[self.SQLFRAME_OUTPUT_DIALECT_KEY]
                if self.SQLFRAME_EXECUTION_DIALECT_KEY in map:
                    self.execution_dialect = map[self.SQLFRAME_EXECUTION_DIALECT_KEY]
                if self.SQLFRAME_CONN_KEY in map:
                    self._session_kwargs["conn"] = map[self.SQLFRAME_CONN_KEY]
                if self.SQLFRAME_SCHEMA_KEY in map:
                    self._session_kwargs["schema"] = map[self.SQLFRAME_SCHEMA_KEY]

        def config(
            self,
            key: t.Optional[str] = None,
            value: t.Optional[t.Any] = None,
            conf: t.Optional[t.Any] = None,
            *,
            map: t.Optional[t.Dict[str, t.Any]] = None,
        ) -> Self:
            self._set_config(key, value, map=map)
            return self

        def _set_session_properties(self) -> None:
            self.session.input_dialect = Dialect.get_or_raise(self.input_dialect)
            self.session.output_dialect = Dialect.get_or_raise(self.output_dialect)
            self.session.execution_dialect = Dialect.get_or_raise(self.execution_dialect)
            if hasattr(self.session, "_connection") and not self.session._connection:
                self.session._connection = self._conn

    builder = Builder()



import pytest

from sqlframe.standalone import functions as F
from sqlframe.standalone.window import Window, WindowSpec


@pytest.mark.parametrize(
    "func, expected",
    [
        (WindowSpec().partitionBy(F.col("cola"), F.col("colb")), "OVER (PARTITION BY cola, colb)"),
        (Window.partitionBy(F.col("cola"), F.col("colb")), "OVER (PARTITION BY cola, colb)"),
        (WindowSpec().orderBy("cola", "colb"), "OVER (ORDER BY cola, colb)"),
        (Window.orderBy("cola", "colb"), "OVER (ORDER BY cola, colb)"),
        (WindowSpec().rowsBetween(3, 5), "OVER (ROWS BETWEEN 3 PRECEDING AND 5 FOLLOWING)"),
        (Window.rowsBetween(3, 5), "OVER (ROWS BETWEEN 3 PRECEDING AND 5 FOLLOWING)"),
        (WindowSpec().rangeBetween(3, 5), "OVER (RANGE BETWEEN 3 PRECEDING AND 5 FOLLOWING)"),
        (Window.rangeBetween(3, 5), "OVER (RANGE BETWEEN 3 PRECEDING AND 5 FOLLOWING)"),
        (
            Window.rowsBetween(Window.unboundedPreceding, 2),
            "OVER (ROWS BETWEEN UNBOUNDED PRECEDING AND 2 FOLLOWING)",
        ),
        (
            Window.rowsBetween(1, Window.unboundedFollowing),
            "OVER (ROWS BETWEEN 1 PRECEDING AND UNBOUNDED FOLLOWING)",
        ),
        (
            Window.rowsBetween(Window.unboundedPreceding, Window.unboundedFollowing),
            "OVER (ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING)",
        ),
        (
            Window.rangeBetween(Window.unboundedPreceding, 2),
            "OVER (RANGE BETWEEN UNBOUNDED PRECEDING AND 2 FOLLOWING)",
        ),
        (
            Window.rangeBetween(1, Window.unboundedFollowing),
            "OVER (RANGE BETWEEN 1 PRECEDING AND UNBOUNDED FOLLOWING)",
        ),
        (
            Window.rangeBetween(Window.unboundedPreceding, Window.unboundedFollowing),
            "OVER (RANGE BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING)",
        ),
    ],
)
def test_window_spec_partition_by(func, expected):
    assert func.sql() == expected



from __future__ import annotations

import typing as t
import warnings

from sqlframe.base.session import _BaseSession
from sqlframe.databricks.catalog import DatabricksCatalog
from sqlframe.databricks.dataframe import DatabricksDataFrame
from sqlframe.databricks.readwriter import (
    DatabricksDataFrameReader,
    DatabricksDataFrameWriter,
)
from sqlframe.databricks.udf import DatabricksUDFRegistration

if t.TYPE_CHECKING:
    from databricks.sql.client import Connection as DatabricksConnection
else:
    DatabricksConnection = t.Any


class DatabricksSession(
    _BaseSession[  # type: ignore
        DatabricksCatalog,
        DatabricksDataFrameReader,
        DatabricksDataFrameWriter,
        DatabricksDataFrame,
        DatabricksConnection,
        DatabricksUDFRegistration,
    ],
):
    _catalog = DatabricksCatalog
    _reader = DatabricksDataFrameReader
    _writer = DatabricksDataFrameWriter
    _df = DatabricksDataFrame
    _udf_registration = DatabricksUDFRegistration

    def __init__(
        self,
        conn: t.Optional[DatabricksConnection] = None,
        server_hostname: t.Optional[str] = None,
        http_path: t.Optional[str] = None,
        access_token: t.Optional[str] = None,
    ):
        from databricks import sql

        if not hasattr(self, "_conn"):
            super().__init__(
                conn or sql.connect(server_hostname, http_path, access_token, disable_pandas=True)
            )

    @classmethod
    def _try_get_map(cls, value: t.Any) -> t.Optional[t.Dict[str, t.Any]]:
        if (
            value
            and isinstance(value, list)
            and all(isinstance(item, tuple) for item in value)
            and all(len(item) == 2 for item in value)
        ):
            return dict(value)
        return None

    class Builder(_BaseSession.Builder):
        DEFAULT_EXECUTION_DIALECT = "databricks"

        @property
        def session(self) -> DatabricksSession:
            return DatabricksSession(**self._session_kwargs)

        def getOrCreate(self) -> DatabricksSession:
            return super().getOrCreate()  # type: ignore

    builder = Builder()



from __future__ import annotations

import logging
import typing as t

from sqlframe.base.catalog import Column as CatalogColumn
from sqlframe.base.dataframe import (
    _BaseDataFrame,
    _BaseDataFrameNaFunctions,
    _BaseDataFrameStatFunctions,
)
from sqlframe.base.mixins.dataframe_mixins import NoCachePersistSupportMixin
from sqlframe.bigquery.group import BigQueryGroupedData

if t.TYPE_CHECKING:
    from sqlframe.bigquery.readwriter import BigQueryDataFrameWriter
    from sqlframe.bigquery.session import BigQuerySession


logger = logging.getLogger(__name__)


class BigQueryDataFrameNaFunctions(_BaseDataFrameNaFunctions["BigQueryDataFrame"]):
    pass


class BigQueryDataFrameStatFunctions(_BaseDataFrameStatFunctions["BigQueryDataFrame"]):
    pass


class BigQueryDataFrame(
    NoCachePersistSupportMixin,
    _BaseDataFrame[
        "BigQuerySession",
        "BigQueryDataFrameWriter",
        "BigQueryDataFrameNaFunctions",
        "BigQueryDataFrameStatFunctions",
        "BigQueryGroupedData",
    ],
):
    _na = BigQueryDataFrameNaFunctions
    _stat = BigQueryDataFrameStatFunctions
    _group_data = BigQueryGroupedData

    @property
    def _typed_columns(self) -> t.List[CatalogColumn]:
        from google.cloud import bigquery

        def field_to_column(field: bigquery.SchemaField) -> CatalogColumn:
            if field.field_type == "RECORD":
                data_type = "STRUCT<"
                for subfield in field.fields:
                    column = field_to_column(subfield)
                    data_type += f"{column.name} {column.dataType},"
                data_type += ">"
            elif field.field_type == "INTEGER":
                data_type = "INT64"
            else:
                data_type = field.field_type
            if field.mode == "REPEATED":
                data_type = f"ARRAY<{data_type}>"
            return CatalogColumn(
                name=field.name,
                dataType=data_type,
                nullable=field.is_nullable,
                description=None,
                isPartition=False,
                isBucket=False,
            )

        job_config = bigquery.QueryJobConfig(dry_run=True, use_query_cache=False)
        sql = self.session._to_sql(self.expression)
        query_job = self.session._client.query(sql, job_config=job_config)
        return [field_to_column(field) for field in query_job.schema]



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import enum
import functools
import itertools
import json
import logging
import sys
import typing as t
import zlib
from copy import copy
from dataclasses import dataclass
from uuid import uuid4

import sqlglot
from prettytable import PrettyTable
from sqlglot import Dialect, maybe_parse
from sqlglot import expressions as exp
from sqlglot import lineage as sqlglot_lineage
from sqlglot.helper import ensure_list, flatten, object_to_dict, seq_get
from sqlglot.optimizer.pushdown_projections import pushdown_projections
from sqlglot.optimizer.qualify import qualify
from sqlglot.optimizer.qualify_columns import quote_identifiers

from sqlframe.base.catalog import Column as CatalogColumn
from sqlframe.base.operations import Operation, operation
from sqlframe.base.transforms import replace_id_value
from sqlframe.base.util import (
    get_func_from_session,
    get_tables_from_expression_with_join,
    normalize_string,
    quote_preserving_alias_or_name,
    sqlglot_to_spark,
    verify_openai_installed,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if t.TYPE_CHECKING:
    import pandas as pd
    from pyarrow import RecordBatchReader
    from pyarrow import Table as ArrowTable
    from sqlglot.dialects.dialect import DialectType

    from sqlframe.base._typing import (
        ColumnOrLiteral,
        ColumnOrName,
        OutputExpressionContainer,
        PrimitiveType,
        StorageLevel,
    )
    from sqlframe.base.column import Column
    from sqlframe.base.group import _BaseGroupedData
    from sqlframe.base.session import WRITER, _BaseSession
    from sqlframe.base.types import Row, StructType

    SESSION = t.TypeVar("SESSION", bound=_BaseSession)
    GROUP_DATA = t.TypeVar("GROUP_DATA", bound=_BaseGroupedData)
else:
    WRITER = t.TypeVar("WRITER")
    SESSION = t.TypeVar("SESSION")
    GROUP_DATA = t.TypeVar("GROUP_DATA")

logger = logging.getLogger(__name__)

JOIN_HINTS = {
    "BROADCAST",
    "BROADCASTJOIN",
    "MAPJOIN",
    "MERGE",
    "SHUFFLEMERGE",
    "MERGEJOIN",
    "SHUFFLE_HASH",
    "SHUFFLE_REPLICATE_NL",
}


DF = t.TypeVar("DF", bound="_BaseDataFrame")


class OpenAIMode(enum.Enum):
    CTE_ONLY = "cte_only"
    FULL = "full"

    @property
    def is_cte_only(self) -> bool:
        return self == OpenAIMode.CTE_ONLY

    @property
    def is_full(self) -> bool:
        return self == OpenAIMode.FULL


@dataclass
class OpenAIConfig:
    mode: OpenAIMode = OpenAIMode.CTE_ONLY
    model: str = "gpt-4o"
    prompt_override: t.Optional[str] = None

    @classmethod
    def from_dict(cls, config: t.Dict[str, t.Any]) -> OpenAIConfig:
        if "mode" in config:
            config["mode"] = OpenAIMode(config["mode"].lower())
        return cls(**config)

    def get_prompt(self, dialect: Dialect) -> str:
        if self.prompt_override:
            return self.prompt_override
        if self.mode.is_cte_only:
            return f"You are a backend tool that creates unique CTE alias names match what a human would write and in snake case. You respond without code blocks and only a json payload with the key being the CTE name that is being replaced and the value being the new CTE human readable name."
        return f"""
        You are a backend tool that converts correct {dialect} SQL to simplified and more human readable version.
        You respond without code block with rewritten {dialect} SQL.
        You don't change any column names in the final select because the user expects those to remain the same.
        You make unique CTE alias names match what a human would write and in snake case.
        You improve formatting with spacing and line-breaks.
        You remove redundant parenthesis and aliases.
        When remove extra quotes, make sure to keep quotes around words that could be reserved words"""


class _BaseDataFrameNaFunctions(t.Generic[DF]):
    def __init__(self, df: DF):
        self.df = df

    def drop(
        self,
        how: str = "any",
        thresh: t.Optional[int] = None,
        subset: t.Optional[t.Union[str, t.Tuple[str, ...], t.List[str]]] = None,
    ) -> DF:
        return self.df.dropna(how=how, thresh=thresh, subset=subset)

    @t.overload
    def fill(self, value: PrimitiveType, subset: t.Optional[t.List[str]] = ...) -> DF: ...

    @t.overload
    def fill(self, value: t.Dict[str, PrimitiveType]) -> DF: ...

    def fill(
        self,
        value: t.Union[PrimitiveType, t.Dict[str, PrimitiveType]],
        subset: t.Optional[t.Union[str, t.Tuple[str, ...], t.List[str]]] = None,
    ) -> DF:
        return self.df.fillna(value=value, subset=subset)

    def replace(
        self,
        to_replace: t.Union[bool, int, float, str, t.List, t.Dict],
        value: t.Optional[t.Union[bool, int, float, str, t.List]] = None,
        subset: t.Optional[t.Union[str, t.List[str]]] = None,
    ) -> DF:
        return self.df.replace(to_replace=to_replace, value=value, subset=subset)


NA = t.TypeVar("NA", bound=_BaseDataFrameNaFunctions)


class _BaseDataFrameStatFunctions(t.Generic[DF]):
    def __init__(self, df: DF):
        self.df = df

    @t.overload
    def approxQuantile(
        self,
        col: str,
        probabilities: t.Union[t.List[float], t.Tuple[float]],
        relativeError: float,
    ) -> t.List[float]: ...

    @t.overload
    def approxQuantile(
        self,
        col: t.Union[t.List[str], t.Tuple[str]],
        probabilities: t.Union[t.List[float], t.Tuple[float]],
        relativeError: float,
    ) -> t.List[t.List[float]]: ...

    def approxQuantile(
        self,
        col: t.Union[str, t.List[str], t.Tuple[str]],
        probabilities: t.Union[t.List[float], t.Tuple[float]],
        relativeError: float,
    ) -> t.Union[t.List[float], t.List[t.List[float]]]:
        return self.df.approxQuantile(col, probabilities, relativeError)

    def corr(self, col1: str, col2: str, method: str = "pearson") -> float:
        return self.df.corr(col1, col2, method)

    def cov(self, col1: str, col2: str) -> float:
        return self.df.cov(col1, col2)


STAT = t.TypeVar("STAT", bound=_BaseDataFrameStatFunctions)


class _BaseDataFrame(t.Generic[SESSION, WRITER, NA, STAT, GROUP_DATA]):
    _na: t.Type[NA]
    _stat: t.Type[STAT]
    _group_data: t.Type[GROUP_DATA]

    def __init__(
        self,
        session: SESSION,
        expression: exp.Select,
        branch_id: t.Optional[str] = None,
        sequence_id: t.Optional[str] = None,
        join_on_uuid: t.Optional[str] = None,
        known_uuids: t.Optional[t.Set[str]] = None,
        last_op: Operation = Operation.INIT,
        pending_hints: t.Optional[t.List[exp.Expression]] = None,
        output_expression_container: t.Optional[OutputExpressionContainer] = None,
        **kwargs,
    ):
        self.session = session
        self.expression: exp.Select = expression
        self.branch_id = branch_id or self.session._random_branch_id
        self.sequence_id = sequence_id or self.session._random_sequence_id
        self.join_on_uuid = join_on_uuid or str(uuid4())
        self.known_uuids = known_uuids or set()
        self.known_uuids.add(self.join_on_uuid)
        self.last_op = last_op
        self.pending_hints = pending_hints or []
        self.output_expression_container = output_expression_container or exp.Select()
        self.temp_views: t.List[exp.Select] = []

    def __getattr__(self, column_name: str) -> Column:
        return self[column_name]

    def __getitem__(self, column_name: str) -> Column:
        from sqlframe.base.util import get_func_from_session

        col_func = get_func_from_session("col", self.session)

        column_name = f"{self.branch_id}.{column_name}"
        col = col_func(column_name)
        col.expression.meta["join_on_uuid"] = self.join_on_uuid
        return col

    def __copy__(self):
        return self.copy()

    @property
    def _typed_columns(self) -> t.List[CatalogColumn]:
        raise NotImplementedError

    @property
    def write(self) -> WRITER:
        return self.session._writer(self)

    @property
    def latest_cte_name(self) -> str:
        if not self.expression.ctes:
            from_exp = self.expression.args["from"]
            if from_exp.alias_or_name:
                return from_exp.alias_or_name
            table_alias = from_exp.find(exp.TableAlias)
            if not table_alias:
                raise RuntimeError(
                    f"Could not find an alias name for this expression: {self.expression}"
                )
            return table_alias.alias_or_name
        return self.expression.ctes[-1].alias

    @property
    def pending_join_hints(self):
        return [hint for hint in self.pending_hints if isinstance(hint, exp.JoinHint)]

    @property
    def pending_partition_hints(self):
        return [hint for hint in self.pending_hints if isinstance(hint, exp.Anonymous)]

    @property
    def columns(self) -> t.List[str]:
        return self.expression.named_selects

    @property
    def na(self) -> NA:
        return self._na(self)

    @property
    def stat(self) -> STAT:
        return self._stat(self)

    @property
    def schema(self) -> StructType:
        """Returns the schema of this :class:`DataFrame` as a :class:`pyspark.sql.types.StructType`.

        .. versionadded:: 1.3.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Returns
        -------
        :class:`StructType`

        Examples
        --------
        >>> df = spark.createDataFrame(
        ...     [(14, "Tom"), (23, "Alice"), (16, "Bob")], ["age", "name"])

        Retrieve the schema of the current DataFrame.

        >>> df.schema
        StructType([StructField('age', LongType(), True),
                    StructField('name', StringType(), True)])
        """
        from sqlframe.base import types

        try:
            return types.StructType(
                [
                    types.StructField(
                        c.name,
                        sqlglot_to_spark(
                            exp.DataType.build(c.dataType, dialect=self.session.output_dialect)
                        ),
                    )
                    for c in self._typed_columns
                ]
            )
        except NotImplementedError as e:
            raise NotImplementedError(
                "This engine does not support schema inference likely since it does not have an active connection."
            ) from e

    def _replace_cte_names_with_hashes(self, expression: exp.Select):
        replacement_mapping = {}
        for cte in expression.ctes:
            old_name_id = cte.args["alias"].this
            new_hashed_id = exp.to_identifier(
                self._create_hash_from_expression(cte.this), quoted=old_name_id.args["quoted"]
            )
            replacement_mapping[old_name_id] = new_hashed_id
            expression = expression.transform(replace_id_value, replacement_mapping).assert_is(
                exp.Select
            )
        return expression

    def _create_cte_from_expression(
        self,
        expression: exp.Expression,
        branch_id: str,
        sequence_id: str,
        name: t.Optional[str] = None,
        **kwargs,
    ) -> t.Tuple[exp.CTE, str]:
        name = name or self._create_hash_from_expression(expression)
        expression_to_cte = expression.copy()
        expression_to_cte.set("with", None)
        cte = exp.Select().with_(name, as_=expression_to_cte, **kwargs).ctes[0]
        cte.set("branch_id", branch_id)
        cte.set("sequence_id", sequence_id)
        return cte, name

    def _ensure_list_of_columns(
        self, cols: t.Optional[t.Union[ColumnOrLiteral, t.Collection[ColumnOrLiteral]]]
    ) -> t.List[Column]:
        from sqlframe.base.column import Column

        return Column.ensure_cols(ensure_list(cols))  # type: ignore

    def _ensure_and_normalize_cols(
        self, cols, expression: t.Optional[exp.Select] = None
    ) -> t.List[Column]:
        from sqlframe.base.normalize import normalize

        cols = self._ensure_list_of_columns(cols)
        normalize(self.session, expression or self.expression, cols)
        return list(flatten([self._expand_star(col) for col in cols]))

    def _ensure_and_normalize_col(self, col):
        from sqlframe.base.column import Column
        from sqlframe.base.normalize import normalize

        col = Column.ensure_col(col)
        normalize(self.session, self.expression, col)
        return col

    def _convert_leaf_to_cte(
        self, sequence_id: t.Optional[str] = None, name: t.Optional[str] = None
    ) -> Self:
        df = self._resolve_pending_hints()
        sequence_id = sequence_id or df.sequence_id
        expression = df.expression.copy()
        cte_expression, cte_name = df._create_cte_from_expression(
            expression=expression, branch_id=self.branch_id, sequence_id=sequence_id, name=name
        )
        new_expression = df._add_ctes_to_expression(
            exp.Select(), expression.ctes + [cte_expression]
        )
        sel_columns = df._get_outer_select_columns(cte_expression)
        new_expression = new_expression.from_(cte_name).select(*[x.expression for x in sel_columns])
        return df.copy(expression=new_expression, sequence_id=sequence_id)

    def _resolve_pending_hints(self) -> Self:
        df = self.copy()
        if not self.pending_hints:
            return df
        expression = df.expression
        hint_expression = expression.args.get("hint") or exp.Hint(expressions=[])
        for hint in df.pending_partition_hints:
            hint_expression.append("expressions", hint)
            df.pending_hints.remove(hint)

        join_aliases = {
            join_table.alias_or_name
            for join_table in get_tables_from_expression_with_join(expression)
        }
        if join_aliases:
            for hint in df.pending_join_hints:
                for sequence_id_expression in hint.expressions:
                    sequence_id_or_name = sequence_id_expression.alias_or_name
                    sequence_ids_to_match = [sequence_id_or_name]
                    if sequence_id_or_name in df.session.name_to_sequence_id_mapping:
                        sequence_ids_to_match = df.session.name_to_sequence_id_mapping[
                            sequence_id_or_name
                        ]
                    matching_ctes = [
                        cte
                        for cte in reversed(expression.ctes)
                        if cte.args["sequence_id"] in sequence_ids_to_match
                    ]
                    for matching_cte in matching_ctes:
                        if matching_cte.alias_or_name in join_aliases:
                            sequence_id_expression.set("this", matching_cte.args["alias"].this)
                            df.pending_hints.remove(hint)
                            break
                hint_expression.append("expressions", hint)
        if hint_expression.expressions:
            expression.set("hint", hint_expression)
        return df

    def _hint(self, hint_name: str, args: t.List[Column]) -> Self:
        hint_name = hint_name.upper()
        hint_expression = (
            exp.JoinHint(
                this=hint_name,
                expressions=[exp.to_table(parameter.alias_or_name) for parameter in args],
            )
            if hint_name in JOIN_HINTS
            else exp.Anonymous(
                this=hint_name, expressions=[parameter.expression for parameter in args]
            )
        )
        new_df = self.copy()
        new_df.pending_hints.append(hint_expression)
        return new_df

    def _set_operation(self, klass: t.Callable, other: Self, distinct: bool) -> Self:
        other_df = other._convert_leaf_to_cte()
        base_expression = self.expression.copy()
        base_expression = self._add_ctes_to_expression(base_expression, other_df.expression.ctes)
        all_ctes = base_expression.ctes
        other_df.expression.set("with", None)
        base_expression.set("with", None)
        operation = klass(this=base_expression, distinct=distinct, expression=other_df.expression)
        operation.set("with", exp.With(expressions=all_ctes))
        return self.copy(expression=operation)._convert_leaf_to_cte()

    def _cache(self, storage_level: str) -> Self:
        df = self._convert_leaf_to_cte()
        df.expression.ctes[-1].set("cache_storage_level", storage_level)
        return df

    def _add_ctes_to_expression(self, expression: exp.Select, ctes: t.List[exp.CTE]) -> exp.Select:
        expression = expression.copy()
        with_expression = expression.args.get("with")
        if with_expression:
            existing_ctes = with_expression.expressions
            existing_cte_counts = {x.alias_or_name: 0 for x in existing_ctes}
            replaced_cte_names = {}  # type: ignore
            for cte in ctes:
                if replaced_cte_names:
                    cte = cte.transform(replace_id_value, replaced_cte_names)  # type: ignore
                if cte.alias_or_name in existing_cte_counts:
                    existing_cte_counts[cte.alias_or_name] += 10
                    cte.set(
                        "this",
                        cte.this.where(
                            exp.EQ(
                                this=exp.Literal.number(existing_cte_counts[cte.alias_or_name]),
                                expression=exp.Literal.number(
                                    existing_cte_counts[cte.alias_or_name]
                                ),
                            )
                        ),
                    )
                    new_cte_alias = self._create_hash_from_expression(cte.this)
                    replaced_cte_names[cte.args["alias"].this] = maybe_parse(
                        new_cte_alias, dialect=self.session.input_dialect, into=exp.Identifier
                    )
                    cte.set(
                        "alias",
                        maybe_parse(
                            new_cte_alias, dialect=self.session.input_dialect, into=exp.TableAlias
                        ),
                    )
                existing_ctes.append(cte)
        else:
            existing_ctes = ctes
        expression.set("with", exp.With(expressions=existing_ctes))
        return expression

    @classmethod
    def _get_outer_select_columns(cls, item: exp.Expression) -> t.List[Column]:
        from sqlframe.base.session import _BaseSession

        col = get_func_from_session("col", _BaseSession())

        outer_select = item.find(exp.Select)
        if outer_select:
            return [col(quote_preserving_alias_or_name(x)) for x in outer_select.expressions]
        return []

    def _create_hash_from_expression(self, expression: exp.Expression) -> str:
        from sqlframe.base.session import _BaseSession

        value = expression.sql(dialect=_BaseSession().input_dialect).encode("utf-8")
        hash = f"t{zlib.crc32(value)}"[:9]
        return self.session._normalize_string(hash)

    def _get_select_expressions(
        self,
    ) -> t.List[t.Tuple[t.Union[t.Type[exp.Cache], OutputExpressionContainer], exp.Select]]:
        select_expressions: t.List[
            t.Tuple[t.Union[t.Type[exp.Cache], OutputExpressionContainer], exp.Select]
        ] = []
        main_select_ctes: t.List[exp.CTE] = []
        for cte in self.expression.ctes:
            cache_storage_level = cte.args.get("cache_storage_level")
            if cache_storage_level:
                select_expression = cte.this.copy()
                select_expression.set("with", exp.With(expressions=copy(main_select_ctes)))
                select_expression.set("cte_alias_name", cte.alias_or_name)
                select_expression.set("cache_storage_level", cache_storage_level)
                select_expressions.append((exp.Cache, select_expression))
            else:
                main_select_ctes.append(cte)
        main_select = self.expression.copy()
        if main_select_ctes:
            main_select.set("with", exp.With(expressions=main_select_ctes))
        expression_select_pair = (type(self.output_expression_container), main_select)
        select_expressions.append(expression_select_pair)  # type: ignore
        return select_expressions

    def _expand_star(self, col: Column) -> t.List[Column]:
        from sqlframe.base.column import Column

        if isinstance(col.column_expression, exp.Star):
            return self._get_outer_select_columns(self.expression)
        elif (
            isinstance(col.column_expression, exp.Column)
            and isinstance(col.column_expression.this, exp.Star)
            and col.column_expression.args.get("table")
        ):
            for cte in self.expression.ctes:
                if cte.alias_or_name == col.column_expression.args["table"].this:
                    return [
                        Column.ensure_col(exp.column(x.column_alias_or_name, cte.alias_or_name))
                        for x in self._get_outer_select_columns(cte)
                    ]
            raise ValueError(
                f"Could not find table to expand star: {col.column_expression.args['table']}"
            )
        return [col]

    def _get_expressions(
        self,
        optimize: bool = True,
        openai_config: t.Optional[t.Union[t.Dict[str, t.Any], OpenAIConfig]] = None,
    ) -> t.List[exp.Expression]:
        df = self._resolve_pending_hints()
        select_expressions = df._get_select_expressions()
        output_expressions: t.List[t.Union[exp.Select, exp.Cache, exp.Drop]] = []
        replacement_mapping: t.Dict[exp.Identifier, exp.Identifier] = {}
        openai_config = (
            OpenAIConfig.from_dict(openai_config)
            if openai_config is not None and isinstance(openai_config, dict)
            else openai_config
        )

        for expression_type, select_expression in select_expressions:
            select_expression = select_expression.transform(
                replace_id_value, replacement_mapping
            ).assert_is(exp.Select)
            if optimize:
                select_expression = t.cast(
                    exp.Select,
                    self.session._optimize(select_expression),
                )
            elif openai_config:
                qualify(
                    select_expression,
                    dialect=self.session.input_dialect,
                    schema=self.session.catalog._schema,
                )
                pushdown_projections(select_expression, schema=self.session.catalog._schema)

            select_expression = df._replace_cte_names_with_hashes(select_expression)

            expression: t.Union[exp.Select, exp.Cache, exp.Drop]
            if expression_type == exp.Cache:
                cache_table_name = df._create_hash_from_expression(select_expression)
                cache_table = exp.to_table(cache_table_name)
                original_alias_name = select_expression.args["cte_alias_name"]

                replacement_mapping[exp.to_identifier(original_alias_name)] = exp.to_identifier(  # type: ignore
                    cache_table_name
                )
                self.session.catalog.add_table(
                    cache_table_name,
                    {
                        quote_preserving_alias_or_name(expression): expression.type.sql(
                            dialect=self.session.input_dialect
                        )
                        if expression.type
                        else "UNKNOWN"
                        for expression in select_expression.expressions
                    },
                )

                cache_storage_level = select_expression.args["cache_storage_level"]
                options = [
                    exp.Literal.string("storageLevel"),
                    exp.Literal.string(cache_storage_level),
                ]
                expression = exp.Cache(
                    this=cache_table, expression=select_expression, lazy=True, options=options
                )

                # We will drop the "view" if it exists before running the cache table
                output_expressions.append(exp.Drop(this=cache_table, exists=True, kind="VIEW"))
            elif expression_type == exp.Create:
                expression = df.output_expression_container.copy()  # type: ignore
                expression.set("expression", select_expression)
            elif expression_type == exp.Insert:
                expression = df.output_expression_container.copy()  # type: ignore
                select_without_ctes = select_expression.copy()
                select_without_ctes.set("with", None)
                expression.set("expression", select_without_ctes)

                if select_expression.ctes:
                    expression.set("with", exp.With(expressions=select_expression.ctes))
            elif expression_type == exp.Select:
                expression = select_expression
            else:
                raise ValueError(f"Invalid expression type: {expression_type}")

            output_expressions.append(expression)
        return output_expressions  # type: ignore

    @t.overload
    def sql(
        self,
        dialect: DialectType = ...,
        optimize: bool = ...,
        pretty: bool = ...,
        *,
        as_list: t.Literal[False],
        **kwargs: t.Any,
    ) -> str: ...

    @t.overload
    def sql(
        self,
        dialect: DialectType = ...,
        optimize: bool = ...,
        pretty: bool = ...,
        *,
        as_list: t.Literal[True],
        **kwargs: t.Any,
    ) -> t.List[str]: ...

    def sql(
        self,
        dialect: DialectType = None,
        optimize: bool = True,
        pretty: bool = True,
        openai_config: t.Optional[t.Union[t.Dict[str, t.Any], OpenAIConfig]] = None,
        as_list: bool = False,
        **kwargs,
    ) -> t.Union[str, t.List[str]]:
        dialect = Dialect.get_or_raise(dialect) if dialect else self.session.output_dialect
        results = []
        for expression in self._get_expressions(optimize=optimize, openai_config=openai_config):
            sql = self.session._to_sql(expression, dialect=dialect, pretty=pretty, **kwargs)
            if openai_config:
                assert isinstance(openai_config, OpenAIConfig)
                verify_openai_installed()
                from openai import OpenAI

                client = OpenAI()
                chat_completed = client.chat.completions.create(
                    messages=[
                        {  # type: ignore
                            "role": "system",
                            "content": openai_config.get_prompt(dialect),
                        },
                        {
                            "role": "user",
                            "content": sql,
                        },
                    ],
                    model=openai_config.model,
                )
                assert chat_completed.choices[0].message.content is not None
                if openai_config.mode.is_cte_only:
                    cte_replacement_mapping = json.loads(chat_completed.choices[0].message.content)
                    for old_name, new_name in cte_replacement_mapping.items():
                        sql = sql.replace(old_name, new_name)
                else:
                    sql = chat_completed.choices[0].message.content
            results.append(sql)

        if as_list:
            return results
        return ";\n".join(results)

    def copy(self, **kwargs) -> Self:
        kwargs["join_on_uuid"] = str(uuid4())
        return self.__class__(**object_to_dict(self, **kwargs))

    @operation(Operation.SELECT)
    def select(self, *cols, **kwargs) -> Self:
        from sqlframe.base.column import Column

        if not cols:
            return self

        if isinstance(cols[0], list):
            cols = cols[0]  # type: ignore
        columns = self._ensure_and_normalize_cols(cols)
        kwargs["append"] = kwargs.get("append", False)
        if self.expression.args.get("joins"):
            ambiguous_cols: t.List[exp.Column] = list(
                flatten(
                    [
                        sub_col
                        for col in columns
                        for sub_col in col.expression.find_all(exp.Column)
                        if not sub_col.table
                    ]
                )
            )
            if ambiguous_cols:
                join_table_identifiers = [
                    x.this for x in get_tables_from_expression_with_join(self.expression)
                ]
                cte_names_in_join = [x.this for x in join_table_identifiers]
                # If we have columns that resolve to multiple CTE expressions then we want to use each CTE left-to-right
                # and therefore we allow multiple columns with the same name in the result. This matches the behavior
                # of Spark.
                resolved_column_position: t.Dict[exp.Column, int] = {
                    col.copy(): -1 for col in ambiguous_cols
                }
                for ambiguous_col in ambiguous_cols:
                    ctes_with_column = [
                        cte
                        for cte in self.expression.ctes
                        if cte.alias_or_name in cte_names_in_join
                        and ambiguous_col.alias_or_name in cte.this.named_selects
                    ]
                    # Check if there is a CTE with this column that we haven't used before. If so, use it. Otherwise,
                    # use the same CTE we used before
                    cte = seq_get(ctes_with_column, resolved_column_position[ambiguous_col] + 1)
                    if cte:
                        resolved_column_position[ambiguous_col] += 1
                    else:
                        cte = ctes_with_column[resolved_column_position[ambiguous_col]]
                    ambiguous_col.set("table", exp.to_identifier(cte.alias_or_name))
        # If an expression is `CAST(x AS DATETYPE)` then we want to alias so that `x` is the result column name
        columns = [
            col.alias(col.expression.alias_or_name)
            if isinstance(col.expression, exp.Cast) and col.expression.alias_or_name
            else col
            for col in columns
        ]
        return self.copy(
            expression=self.expression.select(*[x.expression for x in columns], **kwargs), **kwargs
        )

    @operation(Operation.NO_OP)
    def alias(self, name: str, **kwargs) -> Self:
        from sqlframe.base.column import Column

        new_sequence_id = self.session._random_sequence_id
        df = self.copy()
        for join_hint in df.pending_join_hints:
            for expression in join_hint.expressions:
                if expression.alias_or_name == self.sequence_id:
                    expression.set("this", Column.ensure_col(new_sequence_id).expression)
        df.session._add_alias_to_mapping(name, new_sequence_id)
        return df._convert_leaf_to_cte(sequence_id=new_sequence_id)

    @operation(Operation.WHERE)
    def where(self, column: t.Union[Column, str, bool], **kwargs) -> Self:
        if isinstance(column, str):
            col = self._ensure_and_normalize_col(
                sqlglot.parse_one(column, dialect=self.session.input_dialect)
            )
        else:
            col = self._ensure_and_normalize_col(column)
        return self.copy(expression=self.expression.where(col.expression))

    filter = where

    @operation(Operation.GROUP_BY)
    def groupBy(self, *cols, **kwargs) -> GROUP_DATA:
        columns = self._ensure_and_normalize_cols(cols)
        return self._group_data(self, columns, self.last_op)

    groupby = groupBy

    @operation(Operation.SELECT)
    def agg(self, *exprs, **kwargs) -> Self:
        cols = self._ensure_and_normalize_cols(exprs)
        return self.groupBy().agg(*cols)

    @operation(Operation.FROM)
    def crossJoin(self, other: DF) -> Self:
        """Returns the cartesian product with another :class:`DataFrame`.

        .. versionadded:: 2.1.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        other : :class:`DataFrame`
            Right side of the cartesian product.

        Returns
        -------
        :class:`DataFrame`
            Joined DataFrame.

        Examples
        --------
        >>> from pyspark.sql import Row
        >>> df = spark.createDataFrame(
        ...     [(14, "Tom"), (23, "Alice"), (16, "Bob")], ["age", "name"])
        >>> df2 = spark.createDataFrame(
        ...     [Row(height=80, name="Tom"), Row(height=85, name="Bob")])
        >>> df.crossJoin(df2.select("height")).select("age", "name", "height").show()
        +---+-----+------+
        |age| name|height|
        +---+-----+------+
        | 14|  Tom|    80|
        | 14|  Tom|    85|
        | 23|Alice|    80|
        | 23|Alice|    85|
        | 16|  Bob|    80|
        | 16|  Bob|    85|
        +---+-----+------+
        """
        return self.join.__wrapped__(self, other, how="cross")  # type: ignore

    @operation(Operation.FROM)
    def join(
        self,
        other_df: Self,
        on: t.Optional[t.Union[str, t.List[str], Column, t.List[Column]]] = None,
        how: str = "inner",
        **kwargs,
    ) -> Self:
        from sqlframe.base.functions import coalesce

        if on is None:
            logger.warning("Got no value for on. This appears to change the join to a cross join.")
            how = "cross"

        other_df = other_df._convert_leaf_to_cte()
        join_expression = self._add_ctes_to_expression(self.expression, other_df.expression.ctes)
        # We will determine actual "join on" expression later so we don't provide it at first
        join_expression = join_expression.join(
            join_expression.ctes[-1].alias, join_type=how.replace("_", " ")
        )
        self_columns = self._get_outer_select_columns(join_expression)
        other_columns = self._get_outer_select_columns(other_df.expression)
        join_columns = self._ensure_and_normalize_cols(on)
        # If the two dataframes being joined come from the same branch, we then check if they have any columns that
        # were created using the "branch_id" (df["column_name"]). If so, we know that we need to differentiate
        # the two columns since they would end up with the same table name. We do this by checking for the unique
        # uuids in the other df and finding columns that have metadata on them that match the uuids. If so, we know
        # it comes from the other df and we change the table name to the other df's table name.
        # See `test_self_join` for an example of this.
        if self.branch_id == other_df.branch_id:
            other_df_unique_uuids = other_df.known_uuids - self.known_uuids
            for col in join_columns:
                for col_expr in col.expression.find_all(exp.Column):
                    if (
                        "join_on_uuid" in col_expr.meta
                        and col_expr.meta["join_on_uuid"] in other_df_unique_uuids
                    ):
                        col_expr.set("table", exp.to_identifier(other_df.latest_cte_name))
        # Determines the join clause and select columns to be used passed on what type of columns were provided for
        # the join. The columns returned changes based on how the on expression is provided.
        if how != "cross":
            if isinstance(join_columns[0].expression, exp.Column):
                """
                Unique characteristics of join on column names only:
                * The column names are put at the front of the select list
                * The column names are deduplicated across the entire select list and only the column names (other dups are allowed)
                """
                table_names = [
                    table.alias_or_name
                    for table in get_tables_from_expression_with_join(join_expression)
                ]
                potential_ctes = [
                    cte
                    for cte in join_expression.ctes
                    if cte.alias_or_name in table_names
                    and cte.alias_or_name != other_df.latest_cte_name
                ]
                # Determine the table to reference for the left side of the join by checking each of the left side
                # tables and see if they have the column being referenced.
                join_column_pairs = []
                for join_column in join_columns:
                    num_matching_ctes = 0
                    for cte in potential_ctes:
                        if join_column.alias_or_name in cte.this.named_selects:
                            left_column = join_column.copy().set_table_name(cte.alias_or_name)
                            right_column = join_column.copy().set_table_name(
                                other_df.latest_cte_name
                            )
                            join_column_pairs.append((left_column, right_column))
                            num_matching_ctes += 1
                            # We only want to match one table to the column and that should be matched left -> right
                            # so we break after the first match
                            break
                    if num_matching_ctes == 0:
                        raise ValueError(
                            f"Column `{join_column.alias_or_name}` does not exist in any of the tables."
                        )
                join_clause = functools.reduce(
                    lambda x, y: x & y,
                    [
                        left_column == right_column
                        for left_column, right_column in join_column_pairs
                    ],
                )
                join_column_names = [
                    coalesce(
                        left_col.sql(dialect=self.session.input_dialect),
                        right_col.sql(dialect=self.session.input_dialect),
                    ).alias(left_col.alias_or_name)
                    if how == "full"
                    else left_col.alias_or_name
                    for left_col, right_col in join_column_pairs
                ]
                # To match spark behavior only the join clause gets deduplicated and it gets put in the front of the column list
                select_column_names = [
                    (
                        column.alias_or_name
                        if not isinstance(column.expression.this, exp.Star)
                        else column.sql()
                    )
                    for column in self_columns + other_columns
                ]
                select_column_names = [
                    column_name
                    for column_name in select_column_names
                    if column_name
                    not in [
                        x.alias_or_name if not isinstance(x, str) else x for x in join_column_names
                    ]
                ]
                select_column_names = join_column_names + select_column_names
            else:
                """
                Unique characteristics of join on expressions:
                * There is no deduplication of the results.
                * The left join dataframe columns go first and right come after. No sort preference is given to join columns
                """
                join_columns = self._ensure_and_normalize_cols(join_columns, join_expression)
                if len(join_columns) > 1:
                    join_columns = [functools.reduce(lambda x, y: x & y, join_columns)]
                join_clause = join_columns[0]
                select_column_names = [
                    column.alias_or_name for column in self_columns + other_columns
                ]

            # Update the on expression with the actual join clause to replace the dummy one from before
        else:
            select_column_names = [column.alias_or_name for column in self_columns + other_columns]
            join_clause = None
        join_expression.args["joins"][-1].set("on", join_clause.expression if join_clause else None)
        new_df = self.copy(expression=join_expression)
        new_df.pending_join_hints.extend(self.pending_join_hints)
        new_df.pending_hints.extend(other_df.pending_hints)
        new_df = new_df.select.__wrapped__(new_df, *select_column_names)  # type: ignore
        return new_df

    @operation(Operation.ORDER_BY)
    def orderBy(
        self,
        *cols: t.Union[str, Column],
        ascending: t.Optional[t.Union[t.Any, t.List[t.Any]]] = None,
    ) -> Self:
        """
        This implementation lets any ordered columns take priority over whatever is provided in `ascending`. Spark
        has irregular behavior and can result in runtime errors. Users shouldn't be mixing the two anyways so this
        is unlikely to come up.
        """
        columns = self._ensure_and_normalize_cols(cols)
        pre_ordered_col_indexes = [
            i for i, col in enumerate(columns) if isinstance(col.expression, exp.Ordered)
        ]
        if ascending is None:
            ascending = [True] * len(columns)
        elif not isinstance(ascending, list):
            ascending = [ascending] * len(columns)
        ascending = [bool(x) for i, x in enumerate(ascending)]
        assert len(columns) == len(
            ascending
        ), "The length of items in ascending must equal the number of columns provided"
        col_and_ascending = list(zip(columns, ascending))
        order_by_columns = [
            (
                sqlglot.parse_one(
                    f"{col.expression.sql(dialect=self.session.input_dialect)} {'DESC' if not asc else ''}",
                    dialect=self.session.input_dialect,
                    into=exp.Ordered,
                )
                if i not in pre_ordered_col_indexes
                else columns[i].column_expression
            )
            for i, (col, asc) in enumerate(col_and_ascending)
        ]
        return self.copy(expression=self.expression.order_by(*order_by_columns))

    sort = orderBy

    @operation(Operation.FROM)
    def union(self, other: Self) -> Self:
        return self._set_operation(exp.Union, other, False)

    unionAll = union

    @operation(Operation.FROM)
    def unionByName(self, other: Self, allowMissingColumns: bool = False) -> Self:
        l_columns = self.columns
        r_columns = other.columns
        if not allowMissingColumns:
            l_expressions = l_columns
            r_expressions = l_columns
        else:
            l_expressions = []
            r_expressions = []
            r_columns_unused = copy(r_columns)
            for l_column in l_columns:
                l_expressions.append(l_column)
                if l_column in r_columns:
                    r_expressions.append(l_column)
                    r_columns_unused.remove(l_column)
                else:
                    r_expressions.append(exp.alias_(exp.Null(), l_column, copy=False))
            for r_column in r_columns_unused:
                l_expressions.append(exp.alias_(exp.Null(), r_column, copy=False))
                r_expressions.append(r_column)
        r_df = (
            other.copy()._convert_leaf_to_cte().select(*self._ensure_list_of_columns(r_expressions))
        )
        l_df = self.copy()
        if allowMissingColumns:
            l_df = l_df._convert_leaf_to_cte().select(*self._ensure_list_of_columns(l_expressions))
        return l_df._set_operation(exp.Union, r_df, False)

    @operation(Operation.FROM)
    def intersect(self, other: Self) -> Self:
        return self._set_operation(exp.Intersect, other, True)

    @operation(Operation.FROM)
    def intersectAll(self, other: Self) -> Self:
        return self._set_operation(exp.Intersect, other, False)

    @operation(Operation.FROM)
    def exceptAll(self, other: Self) -> Self:
        return self._set_operation(exp.Except, other, False)

    @operation(Operation.SELECT)
    def distinct(self) -> Self:
        return self.copy(expression=self.expression.distinct())

    @operation(Operation.SELECT)
    def dropDuplicates(self, subset: t.Optional[t.List[str]] = None):
        from sqlframe.base import functions as F
        from sqlframe.base.window import Window

        if not subset:
            return self.distinct()
        column_names = ensure_list(subset)
        window = Window.partitionBy(*column_names).orderBy(*column_names)
        return (
            self.copy()
            .withColumn("row_num", F.row_number().over(window))
            .where(F.col("row_num") == F.lit(1))
            .drop("row_num")
        )

    drop_duplicates = dropDuplicates

    @operation(Operation.FROM)
    def dropna(
        self,
        how: str = "any",
        thresh: t.Optional[int] = None,
        subset: t.Optional[t.Union[str, t.Tuple[str, ...], t.List[str]]] = None,
    ) -> Self:
        from sqlframe.base import functions as F

        minimum_non_null = thresh or 0  # will be determined later if thresh is null
        new_df = self.copy()
        all_columns = self._get_outer_select_columns(new_df.expression)
        if subset:
            null_check_columns = self._ensure_and_normalize_cols(subset)
        else:
            null_check_columns = all_columns
        if thresh is None:
            minimum_num_nulls = 1 if how == "any" else len(null_check_columns)
        else:
            minimum_num_nulls = len(null_check_columns) - minimum_non_null + 1
        if minimum_num_nulls > len(null_check_columns):
            raise RuntimeError(
                f"The minimum num nulls for dropna must be less than or equal to the number of columns. "
                f"Minimum num nulls: {minimum_num_nulls}, Num Columns: {len(null_check_columns)}"
            )
        if_null_checks = [
            F.when(column.isNull(), F.lit(1)).otherwise(F.lit(0)) for column in null_check_columns
        ]
        nulls_added_together = functools.reduce(lambda x, y: x + y, if_null_checks)
        num_nulls = nulls_added_together.alias("num_nulls")
        new_df = new_df.select(num_nulls, append=True)
        filtered_df = new_df.where(F.col("num_nulls") < F.lit(minimum_num_nulls))
        final_df = filtered_df.select(*all_columns)
        return final_df

    def explain(
        self, extended: t.Optional[t.Union[bool, str]] = None, mode: t.Optional[str] = None
    ) -> None:
        """Prints the (logical and physical) plans to the console for debugging purposes.

        .. versionadded:: 1.3.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        extended : bool, optional
            default ``False``. If ``False``, prints only the physical plan.
            When this is a string without specifying the ``mode``, it works as the mode is
            specified.
        mode : str, optional
            specifies the expected output format of plans.

            * ``simple``: Print only a physical plan.
            * ``extended``: Print both logical and physical plans.
            * ``codegen``: Print a physical plan and generated codes if they are available.
            * ``cost``: Print a logical plan and statistics if they are available.
            * ``formatted``: Split explain output into two sections: a physical plan outline \
                and node details.

            .. versionchanged:: 3.0.0
               Added optional argument `mode` to specify the expected output format of plans.

        Examples
        --------
        >>> df = spark.createDataFrame(
        ...     [(14, "Tom"), (23, "Alice"), (16, "Bob")], ["age", "name"])

        Print out the physical plan only (default).

        >>> df.explain()  # doctest: +SKIP
        == Physical Plan ==
        *(1) Scan ExistingRDD[age...,name...]

        Print out all of the parsed, analyzed, optimized and physical plans.

        >>> df.explain(True)
        == Parsed Logical Plan ==
        ...
        == Analyzed Logical Plan ==
        ...
        == Optimized Logical Plan ==
        ...
        == Physical Plan ==
        ...

        Print out the plans with two sections: a physical plan outline and node details

        >>> df.explain(mode="formatted")  # doctest: +SKIP
        == Physical Plan ==
        * Scan ExistingRDD (...)
        (1) Scan ExistingRDD [codegen id : ...]
        Output [2]: [age..., name...]
        ...

        Print a logical plan and statistics if they are available.

        >>> df.explain("cost")
        == Optimized Logical Plan ==
        ...Statistics...
        ...
        """
        sql_queries = self.sql(pretty=False, optimize=False, as_list=True)
        if len(sql_queries) > 1:
            raise ValueError("Cannot explain a DataFrame with multiple queries")
        sql_query = "EXPLAIN " + sql_queries[0]
        self.session._execute(sql_query)

    @operation(Operation.FROM)
    def fillna(
        self,
        value: t.Union[PrimitiveType, t.Dict[str, PrimitiveType]],
        subset: t.Optional[t.Union[str, t.Tuple[str, ...], t.List[str]]] = None,
    ) -> Self:
        """
        Functionality Difference: If you provide a value to replace a null and that type conflicts
        with the type of the column then PySpark will just ignore your replacement.
        This will try to cast them to be the same in some cases. So they won't always match.
        Best to not mix types so make sure replacement is the same type as the column

        Possibility for improvement: Use `typeof` function to get the type of the column
        and check if it matches the type of the value provided. If not then make it null.
        """
        from sqlframe.base import functions as F

        values = None
        columns = None
        new_df = self.copy()
        all_columns = self._get_outer_select_columns(new_df.expression)
        all_column_mapping = {column.alias_or_name: column for column in all_columns}
        if isinstance(value, dict):
            values = list(value.values())
            columns = self._ensure_and_normalize_cols(list(value))
        if not columns:
            columns = self._ensure_and_normalize_cols(subset) if subset else all_columns
        if not values:
            assert not isinstance(value, dict)
            values = [value] * len(columns)
        value_columns = [F.lit(value) for value in values]

        null_replacement_mapping = {
            column.alias_or_name: (
                F.when(column.isNull(), value).otherwise(column).alias(column.alias_or_name)
            )
            for column, value in zip(columns, value_columns)
        }
        null_replacement_mapping = {**all_column_mapping, **null_replacement_mapping}
        null_replacement_columns = [
            null_replacement_mapping[column.alias_or_name] for column in all_columns
        ]
        new_df = new_df.select(*null_replacement_columns)
        return new_df

    @operation(Operation.FROM)
    def replace(
        self,
        to_replace: t.Union[bool, int, float, str, t.List, t.Dict],
        value: t.Optional[t.Union[bool, int, float, str, t.List]] = None,
        subset: t.Optional[t.Collection[ColumnOrName] | ColumnOrName] = None,
    ) -> Self:
        from sqlframe.base import functions as F
        from sqlframe.base.column import Column

        old_values = None
        new_df = self.copy()
        all_columns = self._get_outer_select_columns(new_df.expression)
        all_column_mapping = {column.alias_or_name: column for column in all_columns}

        columns = self._ensure_and_normalize_cols(subset) if subset else all_columns
        if isinstance(to_replace, dict):
            old_values = list(to_replace)
            new_values = list(to_replace.values())
        elif not old_values and isinstance(to_replace, list):
            assert isinstance(value, list), "value must be a list since the replacements are a list"
            assert len(to_replace) == len(
                value
            ), "the replacements and values must be the same length"
            old_values = to_replace
            new_values = value
        else:
            old_values = [to_replace] * len(columns)
            new_values = [value] * len(columns)
        old_values = [F.lit(value) for value in old_values]
        new_values = [F.lit(value) for value in new_values]

        replacement_mapping = {}
        for column in columns:
            # expression = Column(None)
            expression = F.lit(None)
            for i, (old_value, new_value) in enumerate(zip(old_values, new_values)):
                if i == 0:
                    expression = F.when(column == old_value, new_value)
                else:
                    expression = expression.when(column == old_value, new_value)  # type: ignore
            replacement_mapping[column.alias_or_name] = expression.otherwise(column).alias(
                column.expression.alias_or_name
            )

        replacement_mapping = {**all_column_mapping, **replacement_mapping}
        replacement_columns = [replacement_mapping[column.alias_or_name] for column in all_columns]
        new_df = new_df.select(*replacement_columns)
        return new_df

    def transform(self, func: t.Callable[..., DF], *args: t.Any, **kwargs: t.Any) -> Self:
        """Returns a new :class:`DataFrame`. Concise syntax for chaining custom transformations.

        .. versionadded:: 3.0.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        func : function
            a function that takes and returns a :class:`DataFrame`.
        *args
            Positional arguments to pass to func.

            .. versionadded:: 3.3.0
        **kwargs
            Keyword arguments to pass to func.

            .. versionadded:: 3.3.0

        Returns
        -------
        :class:`DataFrame`
            Transformed DataFrame.

        Examples
        --------
        >>> from pyspark.sql.functions import col
        >>> df = spark.createDataFrame([(1, 1.0), (2, 2.0)], ["int", "float"])
        >>> def cast_all_to_int(input_df):
        ...     return input_df.select([col(col_name).cast("int") for col_name in input_df.columns])
        ...
        >>> def sort_columns_asc(input_df):
        ...     return input_df.select(*sorted(input_df.columns))
        ...
        >>> df.transform(cast_all_to_int).transform(sort_columns_asc).show()
        +-----+---+
        |float|int|
        +-----+---+
        |    1|  1|
        |    2|  2|
        +-----+---+

        >>> def add_n(input_df, n):
        ...     return input_df.select([(col(col_name) + n).alias(col_name)
        ...                             for col_name in input_df.columns])
        >>> df.transform(add_n, 1).transform(add_n, n=10).show()
        +---+-----+
        |int|float|
        +---+-----+
        | 12| 12.0|
        | 13| 13.0|
        +---+-----+
        """
        return func(self, *args, **kwargs)  # type: ignore

    @operation(Operation.SELECT)
    def withColumn(self, colName: str, col: Column) -> Self:
        return self.withColumns.__wrapped__(self, {colName: col})  # type: ignore

    @operation(Operation.SELECT)
    def withColumnRenamed(self, existing: str, new: str) -> Self:
        expression = self.expression.copy()
        existing = self.session._normalize_string(existing)
        new = self.session._normalize_string(new)
        existing_columns = [
            expression
            for expression in expression.expressions
            if expression.alias_or_name == existing
        ]
        if not existing_columns:
            raise ValueError("Tried to rename a column that doesn't exist")
        for existing_column in existing_columns:
            if isinstance(existing_column, exp.Column):
                existing_column.replace(exp.alias_(existing_column, new))
            else:
                existing_column.set("alias", exp.to_identifier(new))
        return self.copy(expression=expression)

    @operation(Operation.SELECT)
    def withColumns(self, *colsMap: t.Dict[str, Column]) -> Self:
        """
        Returns a new :class:`DataFrame` by adding multiple columns or replacing the
        existing columns that have the same names.

        The colsMap is a map of column name and column, the column must only refer to attributes
        supplied by this Dataset. It is an error to add columns that refer to some other Dataset.

        .. versionadded:: 3.3.0
           Added support for multiple columns adding

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        colsMap : dict
            a dict of column name and :class:`Column`. Currently, only a single map is supported.

        Returns
        -------
        :class:`DataFrame`
            DataFrame with new or replaced columns.

        Examples
        --------
        >>> df = spark.createDataFrame([(2, "Alice"), (5, "Bob")], schema=["age", "name"])
        >>> df.withColumns({'age2': df.age + 2, 'age3': df.age + 3}).show()
        +---+-----+----+----+
        |age| name|age2|age3|
        +---+-----+----+----+
        |  2|Alice|   4|   5|
        |  5|  Bob|   7|   8|
        +---+-----+----+----+
        """
        if len(colsMap) != 1:
            raise ValueError("Only a single map is supported")
        col_map = {
            self._ensure_and_normalize_col(k).alias_or_name: self._ensure_and_normalize_col(v)
            for k, v in colsMap[0].items()
        }
        existing_cols = self._get_outer_select_columns(self.expression)
        existing_col_names = [x.alias_or_name for x in existing_cols]
        select_columns = existing_cols
        for column_name, col_value in col_map.items():
            existing_col_index = (
                existing_col_names.index(column_name) if column_name in existing_col_names else None
            )
            if existing_col_index is not None:
                select_columns[existing_col_index] = col_value.alias(  # type: ignore
                    column_name
                ).expression
            else:
                select_columns.append(col_value.alias(column_name))
        return self.select.__wrapped__(self, *select_columns)  # type: ignore

    @operation(Operation.SELECT)
    def drop(self, *cols: t.Union[str, Column]) -> Self:
        all_columns = self._get_outer_select_columns(self.expression)
        drop_cols = self._ensure_and_normalize_cols(cols)
        new_columns = [
            col
            for col in all_columns
            if col.alias_or_name not in [drop_column.alias_or_name for drop_column in drop_cols]
        ]
        return self.copy().select(*new_columns, append=False)

    @operation(Operation.LIMIT)
    def limit(self, num: int) -> Self:
        return self.copy(expression=self.expression.limit(num))

    def toDF(self, *cols: str) -> Self:
        """Returns a new :class:`DataFrame` that with new specified column names

        .. versionadded:: 1.6.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        *cols : tuple
            a tuple of string new column name. The length of the
            list needs to be the same as the number of columns in the initial
            :class:`DataFrame`

        Returns
        -------
        :class:`DataFrame`
            DataFrame with new column names.

        Examples
        --------
        >>> df = spark.createDataFrame([(14, "Tom"), (23, "Alice"),
        ...     (16, "Bob")], ["age", "name"])
        >>> df.toDF('f1', 'f2').show()
        +---+-----+
        | f1|   f2|
        +---+-----+
        | 14|  Tom|
        | 23|Alice|
        | 16|  Bob|
        +---+-----+
        """
        if len(cols) != len(self.columns):
            raise ValueError(
                f"Number of column names does not match number of columns: {len(cols)} != {len(self.columns)}"
            )
        expression = self.expression.copy()
        expression = expression.select(
            *[exp.alias_(col, new_col) for col, new_col in zip(expression.expressions, cols)],
            append=False,
        )
        return self.copy(expression=expression)

    @operation(Operation.NO_OP)
    def hint(self, name: str, *parameters: t.Optional[t.Union[str, int]]) -> Self:
        from sqlframe.base.column import Column

        parameter_list = ensure_list(parameters)
        parameter_columns = (
            self._ensure_list_of_columns(parameter_list)
            if parameters
            else Column.ensure_cols([self.sequence_id])
        )
        return self._hint(name, parameter_columns)

    @operation(Operation.NO_OP)
    def repartition(self, numPartitions: t.Union[int, ColumnOrName], *cols: ColumnOrName) -> Self:
        num_partition_cols = self._ensure_list_of_columns(numPartitions)
        columns = self._ensure_and_normalize_cols(cols)
        args = num_partition_cols + columns
        return self._hint("repartition", args)

    @operation(Operation.NO_OP)
    def coalesce(self, numPartitions: int) -> Self:
        lit = get_func_from_session("lit")

        num_partitions = lit(numPartitions)
        return self._hint("coalesce", [num_partitions])

    @operation(Operation.NO_OP)
    def cache(self) -> Self:
        return self._cache(storage_level="MEMORY_AND_DISK")

    @operation(Operation.NO_OP)
    def persist(self, storageLevel: StorageLevel = "MEMORY_AND_DISK_SER") -> Self:
        """
        Storage Level Options: https://spark.apache.org/docs/3.0.0-preview/sql-ref-syntax-aux-cache-cache-table.html
        """
        return self._cache(storageLevel)

    @t.overload
    def cube(self, *cols: ColumnOrName) -> GROUP_DATA: ...

    @t.overload
    def cube(self, __cols: t.Union[t.List[Column], t.List[str]]) -> GROUP_DATA: ...

    def cube(self, *cols: ColumnOrName) -> GROUP_DATA:  # type: ignore[misc]
        """
        Create a multi-dimensional cube for the current :class:`DataFrame` using
        the specified columns, so we can run aggregations on them.

        .. versionadded:: 1.4.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        cols : list, str or :class:`Column`
            columns to create cube by.
            Each element should be a column name (string) or an expression (:class:`Column`)
            or list of them.

        Returns
        -------
        :class:`GroupedData`
            Cube of the data by given columns.

        Examples
        --------
        >>> df = spark.createDataFrame([(2, "Alice"), (5, "Bob")], schema=["age", "name"])
        >>> df.cube("name", df.age).count().orderBy("name", "age").show()
        +-----+----+-----+
        | name| age|count|
        +-----+----+-----+
        | NULL|NULL|    2|
        | NULL|   2|    1|
        | NULL|   5|    1|
        |Alice|NULL|    1|
        |Alice|   2|    1|
        |  Bob|NULL|    1|
        |  Bob|   5|    1|
        +-----+----+-----+
        """

        columns = self._ensure_and_normalize_cols(cols)
        grouping_columns: t.List[t.List[Column]] = []
        for i in reversed(range(len(columns) + 1)):
            grouping_columns.extend([list(x) for x in itertools.combinations(columns, i)])
        return self._group_data(self, grouping_columns, self.last_op)

    def collect(self) -> t.List[Row]:
        return self._collect()

    def _collect(self, **kwargs) -> t.List[Row]:
        return self.session._collect(self._get_expressions(optimize=False), **kwargs)

    @t.overload
    def head(self) -> t.Optional[Row]: ...

    @t.overload
    def head(self, n: int) -> t.List[Row]: ...

    def head(self, n: t.Optional[int] = None) -> t.Union[t.Optional[Row], t.List[Row]]:
        n = n or 1
        df = self.limit(n)
        if n == 1:
            return df.collect()[0]
        return df.collect()

    def first(self) -> t.Optional[Row]:
        return self.head()

    def show(
        self, n: int = 20, truncate: t.Optional[t.Union[bool, int]] = None, vertical: bool = False
    ):
        if vertical:
            raise NotImplementedError("Vertical show is not yet supported")
        if truncate:
            logger.warning("Truncate is ignored so full results will be displayed")
        result = self.limit(n).collect()
        table = PrettyTable()
        if row := seq_get(result, 0):
            table.field_names = row._unique_field_names
            for row in result:
                table.add_row(list(row))
        print(table)

    def printSchema(self, level: t.Optional[int] = None) -> None:
        def print_schema(
            column_name: str, column_type: exp.DataType, nullable: bool, current_level: int
        ):
            if level and current_level >= level:
                return
            if current_level > 0:
                print(" |   " * current_level, end="")
            print(
                f" |-- {column_name}: {column_type.sql(self.session.output_dialect).lower()} (nullable = {str(nullable).lower()})"
            )
            if column_type.this in (exp.DataType.Type.STRUCT, exp.DataType.Type.OBJECT):
                for column_def in column_type.expressions:
                    print_schema(column_def.name, column_def.args["kind"], True, current_level + 1)
            if column_type.this == exp.DataType.Type.ARRAY:
                for data_type in column_type.expressions:
                    print_schema("element", data_type, True, current_level + 1)
            if column_type.this == exp.DataType.Type.MAP:
                print_schema("key", column_type.expressions[0], True, current_level + 1)
                print_schema("value", column_type.expressions[1], True, current_level + 1)

        print("root")
        for column in self._typed_columns:
            print_schema(
                column.name,
                exp.DataType.build(column.dataType, dialect=self.session.output_dialect),
                column.nullable,
                0,
            )

    def lineage(self, col: ColumnOrName, optimize: bool = True) -> sqlglot_lineage.Node:
        return sqlglot_lineage.lineage(
            column=self._ensure_and_normalize_col(col).alias_or_name,
            sql=self._get_expressions(optimize=optimize)[0],
            schema=self.session.catalog._schema,
        )

    def toPandas(self) -> pd.DataFrame:
        return self.session._fetchdf(self._get_expressions(optimize=False))

    def createOrReplaceTempView(self, name: str) -> None:
        name = normalize_string(name, from_dialect="input")
        df = self.copy()._convert_leaf_to_cte()
        self.session.temp_views[name] = df
        self.session.catalog.add_table(
            name, [x.alias_or_name for x in self._get_outer_select_columns(df.expression)]
        )

    def count(self) -> int:
        if not self.session._has_connection:
            raise RuntimeError("Cannot count without a connection")

        df = self._convert_leaf_to_cte()
        df = self.copy(expression=df.expression.select("count(*)", append=False))
        return df.collect()[0][0]

    def createGlobalTempView(self, name: str) -> None:
        raise NotImplementedError("Global temp views are not yet supported")

    """
    Stat Functions
    """

    @t.overload
    def approxQuantile(
        self,
        col: str,
        probabilities: t.Union[t.List[float], t.Tuple[float]],
        relativeError: float,
    ) -> t.List[float]: ...

    @t.overload
    def approxQuantile(
        self,
        col: t.Union[t.List[str], t.Tuple[str]],
        probabilities: t.Union[t.List[float], t.Tuple[float]],
        relativeError: float,
    ) -> t.List[t.List[float]]: ...

    def approxQuantile(
        self,
        col: t.Union[str, t.List[str], t.Tuple[str]],
        probabilities: t.Union[t.List[float], t.Tuple[float]],
        relativeError: float,
    ) -> t.Union[t.List[float], t.List[t.List[float]]]:
        """
        Calculates the approximate quantiles of numerical columns of a
        :class:`DataFrame`.

        The result of this algorithm has the following deterministic bound:
        If the :class:`DataFrame` has N elements and if we request the quantile at
        probability `p` up to error `err`, then the algorithm will return
        a sample `x` from the :class:`DataFrame` so that the *exact* rank of `x` is
        close to (p * N). More precisely,

          floor((p - err) * N) <= rank(x) <= ceil((p + err) * N).

        This method implements a variation of the Greenwald-Khanna
        algorithm (with some speed optimizations). The algorithm was first
        present in [[https://doi.org/10.1145/375663.375670
        Space-efficient Online Computation of Quantile Summaries]]
        by Greenwald and Khanna.

        .. versionadded:: 2.0.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        col: str, tuple or list
            Can be a single column name, or a list of names for multiple columns.

            .. versionchanged:: 2.2.0
               Added support for multiple columns.
        probabilities : list or tuple
            a list of quantile probabilities
            Each number must belong to [0, 1].
            For example 0 is the minimum, 0.5 is the median, 1 is the maximum.
        relativeError : float
            The relative target precision to achieve
            (>= 0). If set to zero, the exact quantiles are computed, which
            could be very expensive. Note that values greater than 1 are
            accepted but gives the same result as 1.

        Returns
        -------
        list
            the approximate quantiles at the given probabilities.

            * If the input `col` is a string, the output is a list of floats.

            * If the input `col` is a list or tuple of strings, the output is also a
                list, but each element in it is a list of floats, i.e., the output
                is a list of list of floats.

        Notes
        -----
        Null values will be ignored in numerical columns before calculation.
        For columns only containing null values, an empty list is returned.
        """

        percentile_approx = get_func_from_session("percentile_approx")
        col_func = get_func_from_session("col")

        accuracy = 1.0 / relativeError if relativeError > 0.0 else 10000

        df = self.select(
            *[
                percentile_approx(col_func(x), probabilities, accuracy).alias(f"val_{i}")
                for i, x in enumerate(ensure_list(col))
            ]
        )
        rows = df.collect()
        return [[float(y) for y in x] for row in rows for x in row.asDict().values()]

    def corr(self, col1: str, col2: str, method: t.Optional[str] = None) -> float:
        """
        Calculates the correlation of two columns of a :class:`DataFrame` as a double value.
        Currently only supports the Pearson Correlation Coefficient.
        :func:`DataFrame.corr` and :func:`DataFrameStatFunctions.corr` are aliases of each other.

        .. versionadded:: 1.4.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        col1 : str
            The name of the first column
        col2 : str
            The name of the second column
        method : str, optional
            The correlation method. Currently only supports "pearson"

        Returns
        -------
        float
            Pearson Correlation Coefficient of two columns.

        Examples
        --------
        >>> df = spark.createDataFrame([(1, 12), (10, 1), (19, 8)], ["c1", "c2"])
        >>> df.corr("c1", "c2")
        -0.3592106040535498
        >>> df = spark.createDataFrame([(11, 12), (10, 11), (9, 10)], ["small", "bigger"])
        >>> df.corr("small", "bigger")
        1.0
        """
        if method != "pearson":
            raise ValueError(f"Currently only the Pearson Correlation Coefficient is supported")

        corr = get_func_from_session("corr")
        col_func = get_func_from_session("col")

        return self.select(corr(col_func(col1), col_func(col2))).collect()[0][0]

    def cov(self, col1: str, col2: str) -> float:
        """
        Calculate the sample covariance for the given columns, specified by their names, as a
        double value. :func:`DataFrame.cov` and :func:`DataFrameStatFunctions.cov` are aliases.

        .. versionadded:: 1.4.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        col1 : str
            The name of the first column
        col2 : str
            The name of the second column

        Returns
        -------
        float
            Covariance of two columns.

        Examples
        --------
        >>> df = spark.createDataFrame([(1, 12), (10, 1), (19, 8)], ["c1", "c2"])
        >>> df.cov("c1", "c2")
        -18.0
        >>> df = spark.createDataFrame([(11, 12), (10, 11), (9, 10)], ["small", "bigger"])
        >>> df.cov("small", "bigger")
        1.0

        """
        covar_samp = get_func_from_session("covar_samp")
        col_func = get_func_from_session("col")

        return self.select(covar_samp(col_func(col1), col_func(col2))).collect()[0][0]

    @t.overload
    def toArrow(self) -> ArrowTable: ...

    @t.overload
    def toArrow(self, batch_size: int) -> RecordBatchReader: ...

    def toArrow(self, batch_size: t.Optional[int] = None) -> t.Union[ArrowTable, RecordBatchReader]:
        """
        `batch_size` and `RecordBatchReader` are not part of the PySpark API
        """
        raise NotImplementedError("Arrow conversion is not supported by this engine")



import fnmatch
import typing as t

from sqlglot import exp

from sqlframe.base.catalog import (
    DF,
    SESSION,
    CatalogMetadata,
    Column,
    Database,
    Table,
    _BaseCatalog,
)
from sqlframe.base.util import normalize_string, schema_, to_schema


class _BaseInfoSchemaMixin(_BaseCatalog, t.Generic[SESSION, DF]):
    QUALIFY_INFO_SCHEMA_WITH_DATABASE = False
    UPPERCASE_INFO_SCHEMA = False

    def _get_info_schema_table(
        self,
        table_name: str,
        database: t.Optional[str] = None,
        qualify_override: t.Optional[bool] = None,
    ) -> exp.Table:
        table = table_name
        db = "information_schema"
        if self.UPPERCASE_INFO_SCHEMA:
            table = table.upper()
            db = db.upper()
        qualify = (
            qualify_override
            if qualify_override is not None
            else self.QUALIFY_INFO_SCHEMA_WITH_DATABASE
        )
        if qualify:
            if database:
                catalog = normalize_string(database, from_dialect="input", to_dialect="output")
            else:
                catalog = self.currentDatabase()
            if not db:
                raise ValueError("Table name must be qualified with a database.")
            return exp.table_(
                catalog=exp.to_identifier(catalog, quoted=True),
                db=exp.to_identifier(db, quoted=True),
                table=exp.to_identifier(table, quoted=True),
            )
        return exp.table_(
            db=exp.to_identifier(db, quoted=True), table=exp.to_identifier(table, quoted=True)
        )


class GetCurrentCatalogFromFunctionMixin(_BaseCatalog, t.Generic[SESSION, DF]):
    CURRENT_CATALOG_EXPRESSION: exp.Expression = exp.func("current_catalog")

    def currentCatalog(self) -> str:
        """Returns the current default catalog in this session.

        .. versionadded:: 3.4.0

        Examples
        --------
        >>> spark.catalog.currentCatalog()
        'spark_catalog'
        """
        return normalize_string(
            self.session._collect(
                exp.select(self.CURRENT_CATALOG_EXPRESSION), quote_identifiers=False
            )[0][0],
            from_dialect="execution",
            to_dialect="output",
        )


class GetCurrentDatabaseFromFunctionMixin(_BaseCatalog, t.Generic[SESSION, DF]):
    CURRENT_DATABASE_EXPRESSION: exp.Expression = exp.func("current_schema")

    def currentDatabase(self) -> str:
        """Returns the current default schema in this session.

        .. versionadded:: 3.4.0

        Examples
        --------
        >>> spark.catalog.currentDatabase()
        'default'
        """
        return normalize_string(
            self.session._collect(exp.select(self.CURRENT_DATABASE_EXPRESSION))[0][0],
            from_dialect="execution",
            to_dialect="output",
        )


class SetCurrentCatalogFromUseMixin(_BaseCatalog, t.Generic[SESSION, DF]):
    def setCurrentCatalog(self, catalogName: str) -> None:
        """Sets the current default catalog in this session.

        .. versionadded:: 3.4.0

        Parameters
        ----------
        catalogName : str
            name of the catalog to set

        Examples
        --------
        >>> spark.catalog.setCurrentCatalog("spark_catalog")
        """
        self.session._collect(
            exp.Use(this=exp.parse_identifier(catalogName, dialect=self.session.input_dialect))
        )


class ListDatabasesFromInfoSchemaMixin(_BaseInfoSchemaMixin, t.Generic[SESSION, DF]):
    def listDatabases(self, pattern: t.Optional[str] = None) -> t.List[Database]:
        """
        Returns a t.List of databases available across all sessions.

        .. versionadded:: 2.0.0

        Parameters
        ----------
        pattern : str
            The pattern that the database name needs to match.

            .. versionchanged: 3.5.0
                Adds ``pattern`` argument.

        Returns
        -------
        t.List
            A t.List of :class:`Database`.

        Examples
        --------
        >>> spark.catalog.t.listDatabases()
        [Database(name='default', catalog='spark_catalog', description='default database', ...

        >>> spark.catalog.t.listDatabases("def*")
        [Database(name='default', catalog='spark_catalog', description='default database', ...

        >>> spark.catalog.t.listDatabases("def2*")
        []
        """
        table = self._get_info_schema_table("schemata", qualify_override=False)
        results = self.session._collect(
            # The table name can be specifically formatted in uppercase to run properly and we don't want to
            # normalize that away
            exp.Select().select("schema_name", "catalog_name").from_(table),
            skip_normalization=True,
        )
        databases = [
            Database(
                name=normalize_string(x[0], from_dialect="execution", to_dialect="output"),
                catalog=normalize_string(x[1], from_dialect="execution", to_dialect="output"),
                description=None,
                locationUri="",
            )
            for x in results
        ]
        if pattern:
            normalized_pattern = normalize_string(
                pattern, from_dialect="input", to_dialect="output", is_pattern=True
            )
            databases = [db for db in databases if fnmatch.fnmatch(db.name, normalized_pattern)]
        return databases


class ListCatalogsFromInfoSchemaMixin(_BaseInfoSchemaMixin, t.Generic[SESSION, DF]):
    def listCatalogs(self, pattern: t.Optional[str] = None) -> t.List[CatalogMetadata]:
        """
        Returns a t.List of databases available across all sessions.

        .. versionadded:: 2.0.0

        Parameters
        ----------
        pattern : str
            The pattern that the database name needs to match.

            .. versionchanged: 3.5.0
                Adds ``pattern`` argument.

        Returns
        -------
        t.List
            A t.List of :class:`Database`.

        Examples
        --------
        >>> spark.catalog.t.listDatabases()
        [Database(name='default', catalog='spark_catalog', description='default database', ...

        >>> spark.catalog.t.listDatabases("def*")
        [Database(name='default', catalog='spark_catalog', description='default database', ...

        >>> spark.catalog.t.listDatabases("def2*")
        []
        """
        table = self._get_info_schema_table("schemata")
        results = self.session._collect(
            exp.Select().select("catalog_name").from_(table).distinct(), skip_normalization=True
        )
        catalogs = [
            CatalogMetadata(
                name=normalize_string(x[0], from_dialect="execution", to_dialect="output"),
                description=None,
            )
            for x in results
        ]
        if pattern:
            normalized_pattern = normalize_string(
                pattern, from_dialect="input", to_dialect="output", is_pattern=True
            )
            catalogs = [
                catalog for catalog in catalogs if fnmatch.fnmatch(catalog.name, normalized_pattern)
            ]
        return catalogs


class SetCurrentDatabaseFromSearchPathMixin(_BaseCatalog, t.Generic[SESSION, DF]):
    def setCurrentDatabase(self, dbName: str) -> None:
        """
        Sets the current default database in this session.

        .. versionadded:: 2.0.0

        Examples
        --------
        >>> spark.catalog.setCurrentDatabase("default")
        """
        self.session._execute(f'SET search_path TO "{dbName}"')


class SetCurrentDatabaseFromUseMixin(_BaseCatalog, t.Generic[SESSION, DF]):
    def setCurrentDatabase(self, dbName: str) -> None:
        """
        Sets the current default database in this session.

        .. versionadded:: 2.0.0

        Examples
        --------
        >>> spark.catalog.setCurrentDatabase("default")
        """
        dbName = normalize_string(dbName, from_dialect="input", to_dialect="output", is_schema=True)
        schema = to_schema(dbName, dialect=self.session.output_dialect)

        if not schema.catalog:
            schema.set(
                "catalog",
                exp.parse_identifier(self.currentCatalog(), dialect=self.session.output_dialect),
            )
        self.session._collect(exp.Use(this=schema))


class ListTablesFromInfoSchemaMixin(_BaseInfoSchemaMixin, t.Generic[SESSION, DF]):
    def listTables(
        self, dbName: t.Optional[str] = None, pattern: t.Optional[str] = None
    ) -> t.List[Table]:
        """Returns a t.List of tables/views in the specified database.

        .. versionadded:: 2.0.0

        Parameters
        ----------
        dbName : str
            name of the database to t.List the tables.

            .. versionchanged:: 3.4.0
               Allow ``dbName`` to be qualified with catalog name.

        pattern : str
            The pattern that the database name needs to match.

            .. versionchanged: 3.5.0
                Adds ``pattern`` argument.

        Returns
        -------
        t.List
            A t.List of :class:`Table`.

        Notes
        -----
        If no database is specified, the current database and catalog
        are used. This API includes all temporary views.

        Examples
        --------
        >>> spark.range(1).createTempView("test_view")
        >>> spark.catalog.t.listTables()
        [Table(name='test_view', catalog=None, namespace=[], description=None, ...

        >>> spark.catalog.t.listTables(pattern="test*")
        [Table(name='test_view', catalog=None, namespace=[], description=None, ...

        >>> spark.catalog.t.listTables(pattern="table*")
        []

        >>> _ = spark.catalog.dropTempView("test_view")
        >>> spark.catalog.t.listTables()
        []
        """
        if dbName is None and pattern is None:
            database = normalize_string(
                self.currentDatabase(), from_dialect="output", to_dialect="input"
            )
            catalog = normalize_string(
                self.currentCatalog(), from_dialect="output", to_dialect="input"
            )
            schema = schema_(
                db=exp.parse_identifier(database, dialect=self.session.input_dialect),
                catalog=exp.parse_identifier(catalog, dialect=self.session.input_dialect),
            )
        elif dbName:
            dbName = normalize_string(dbName, from_dialect="input", is_schema=True)
            schema = to_schema(dbName, dialect=self.session.input_dialect)
        else:
            schema = None
        table = self._get_info_schema_table("tables", database=schema.db if schema else None)
        select = exp.select(
            'table_name AS "table_name"',
            'table_schema AS "table_schema"',
            'table_catalog AS "table_catalog"',
            'table_type AS "table_type"',
        ).from_(table)
        if schema and schema.db:
            select = select.where(
                exp.column("table_schema").eq(
                    normalize_string(
                        schema.db,
                        from_dialect="input",
                        to_dialect="execution",
                        to_string_literal=True,
                    )
                )
            )
        if schema and schema.catalog:
            select = select.where(
                exp.column("table_catalog").eq(
                    normalize_string(
                        schema.catalog,
                        from_dialect="input",
                        to_dialect="execution",
                        to_string_literal=True,
                    )
                )
            )
        results = self.session._collect(select, skip_normalization=True)
        tables = [
            Table(
                name=normalize_string(
                    x["table_name"], from_dialect="execution", to_dialect="output"
                ),
                catalog=normalize_string(
                    x["table_catalog"], from_dialect="execution", to_dialect="output"
                ),
                namespace=[
                    normalize_string(
                        x["table_schema"], from_dialect="execution", to_dialect="output"
                    )
                ],
                description=None,
                tableType="VIEW" if x["table_type"] == "VIEW" else "MANAGED",
                isTemporary=False,
            )
            for x in results
        ]
        for table in self.session.temp_views.keys():
            tables.append(
                Table(
                    name=table,  # type: ignore
                    catalog=None,
                    namespace=[],
                    description=None,
                    tableType="VIEW",
                    isTemporary=True,
                )
            )
        if pattern:
            tables = [
                x
                for x in tables
                if fnmatch.fnmatch(
                    x.name,
                    normalize_string(
                        pattern, from_dialect="input", to_dialect="output", is_pattern=True
                    ),
                )
            ]
        return tables


class ListColumnsFromInfoSchemaMixin(_BaseInfoSchemaMixin, t.Generic[SESSION, DF]):
    def listColumns(
        self, tableName: str, dbName: t.Optional[str] = None, include_temp: bool = False
    ) -> t.List[Column]:
        """Returns a t.List of columns for the given table/view in the specified database.

        .. versionadded:: 2.0.0

        Parameters
        ----------
        tableName : str
            name of the table to t.List columns.

            .. versionchanged:: 3.4.0
               Allow ``tableName`` to be qualified with catalog name when ``dbName`` is None.

        dbName : str, t.Optional
            name of the database to find the table to t.List columns.

        Returns
        -------
        t.List
            A t.List of :class:`Column`.

        Notes
        -----
        The order of arguments here is different from that of its JVM counterpart
        because Python does not support method overloading.

        If no database is specified, the current database and catalog
        are used. This API includes all temporary views.

        Examples
        --------
        >>> _ = spark.sql("DROP TABLE IF EXISTS tbl1")
        >>> _ = spark.sql("CREATE TABLE tblA (name STRING, age INT) USING parquet")
        >>> spark.catalog.t.listColumns("tblA")
        [Column(name='name', description=None, dataType='string', nullable=True, ...
        >>> _ = spark.sql("DROP TABLE tblA")
        """
        tableName = normalize_string(tableName, from_dialect="input", is_table=True)
        dbName = normalize_string(dbName, from_dialect="input", is_schema=True) if dbName else None
        if df := self.session.temp_views.get(tableName):
            return [
                Column(
                    name=x,
                    description=None,
                    dataType="",
                    nullable=True,
                    isPartition=False,
                    isBucket=False,
                )
                for x in df.columns
            ]

        table = exp.to_table(tableName, dialect=self.session.input_dialect)
        schema = to_schema(dbName, dialect=self.session.input_dialect) if dbName else None
        if not table.db:
            if schema and schema.db:
                table.set("db", schema.args["db"])
            else:
                current_database = normalize_string(
                    self.currentDatabase(), from_dialect="output", to_dialect="input"
                )
                table.set(
                    "db",
                    exp.parse_identifier(current_database, dialect=self.session.input_dialect),
                )
        if not table.catalog:
            if schema and schema.catalog:
                table.set("catalog", schema.args["catalog"])
            else:
                current_catalog = normalize_string(
                    self.currentCatalog(), from_dialect="output", to_dialect="input"
                )
                table.set(
                    "catalog",
                    exp.parse_identifier(current_catalog, dialect=self.session.input_dialect),
                )
        source_table = self._get_info_schema_table("columns", database=table.db)
        select = (
            exp.select(
                'column_name AS "column_name"',
                'data_type AS "data_type"',
                'is_nullable AS "is_nullable"',
            )
            .from_(source_table)
            .where(
                exp.column("table_name").eq(
                    normalize_string(
                        table.name,
                        from_dialect="input",
                        to_dialect="execution",
                        to_string_literal=True,
                    )
                )
            )
        )
        if table.db:
            schema_filter: exp.Expression = exp.column("table_schema").eq(
                normalize_string(
                    table.db, from_dialect="input", to_dialect="execution", to_string_literal=True
                )
            )
            if include_temp and self.TEMP_SCHEMA_FILTER:
                schema_filter = exp.Or(this=schema_filter, expression=self.TEMP_SCHEMA_FILTER)
            select = select.where(schema_filter)
        if table.catalog:
            catalog_filter: exp.Expression = exp.column("table_catalog").eq(
                normalize_string(
                    table.catalog,
                    from_dialect="input",
                    to_dialect="execution",
                    to_string_literal=True,
                )
            )
            if include_temp and self.TEMP_CATALOG_FILTER:
                catalog_filter = exp.Or(this=catalog_filter, expression=self.TEMP_CATALOG_FILTER)
            select = select.where(catalog_filter)
        results = self.session._collect(select, skip_normalization=True)
        return [
            Column(
                name=normalize_string(
                    x["column_name"], from_dialect="execution", to_dialect="output"
                ),
                description=None,
                dataType=normalize_string(
                    x["data_type"], from_dialect="execution", to_dialect="output", is_datatype=True
                ),
                nullable=x["is_nullable"] == "YES",
                isPartition=False,
                isBucket=False,
            )
            for x in results
        ]



from sqlframe.base.function_alternatives import (  # noqa
    any_value_ignore_nulls_not_supported as any_value,
    e_literal as e,
    expm1_from_exp as expm1,
    log1p_from_log as log1p,
    rint_from_round as rint,
    collect_set_from_list_distinct as collect_set,
    isnan_using_equal as isnan,
    isnull_using_equal as isnull,
    nanvl_as_case as nanvl,
    rand_no_seed as rand,
    round_cast_as_numeric as round,
    year_from_extract as year,
    quarter_from_extract as quarter,
    month_from_extract as month,
    dayofweek_from_extract_with_isodow as dayofweek,
    dayofmonth_from_extract_with_day as dayofmonth,
    dayofyear_from_extract_doy as dayofyear,
    hour_from_extract as hour,
    minute_from_extract as minute,
    second_from_extract as second,
    weekofyear_from_extract_as_week as weekofyear,
    make_date_casted_as_integer as make_date,
    date_add_by_multiplication as date_add,
    date_sub_by_multiplication as date_sub,
    date_diff_with_subtraction as date_diff,
    date_diff_with_subtraction as datediff,
    add_months_by_multiplication as add_months,
    months_between_from_age_and_extract as months_between,
    from_unixtime_from_timestamp as from_unixtime,
    unix_timestamp_from_extract as unix_timestamp,
    base64_from_blob as base64,
    bas64_from_encode as base64,
    unbase64_from_decode as unbase64,
    decode_from_convert_from as decode,
    encode_from_convert_to as encode,
    format_number_from_to_char as format_number,
    format_string_with_format as format_string,
    split_from_regex_split_to_array as split,
    array_contains_any as array_contains,
    slice_with_brackets as slice,
    element_at_using_brackets as element_at,
    get_json_object_using_arrow_op as get_json_object,
    array_min_from_subquery as array_min,
    array_max_from_subquery as array_max,
    left_cast_len as left,
    right_cast_len as right,
    position_cast_start as position,
    try_element_at_zero_based as try_element_at,
    regexp_replace_global_option as regexp_replace,
)
from sqlframe.base.functions import (
    abs as abs,
    acos as acos,
    acosh as acosh,
    array as array,
    array_join as array_join,
    array_position as array_position,
    array_remove as array_remove,
    arrays_overlap as arrays_overlap,
    asc as asc,
    asc_nulls_first as asc_nulls_first,
    asc_nulls_last as asc_nulls_last,
    ascii as ascii,
    asin as asin,
    asinh as asinh,
    atan as atan,
    atan2 as atan2,
    atanh as atanh,
    avg as avg,
    bit_length as bit_length,
    bitwiseNOT as bitwiseNOT,
    bitwise_not as bitwise_not,
    bool_and as bool_and,
    bool_or as bool_or,
    call_function as call_function,
    cbrt as cbrt,
    ceil as ceil,
    ceiling as ceiling,
    char as char,
    coalesce as coalesce,
    col as col,
    collect_list as collect_list,
    concat as concat,
    concat_ws as concat_ws,
    corr as corr,
    cos as cos,
    cosh as cosh,
    cot as cot,
    count as count,
    countDistinct as countDistinct,
    count_distinct as count_distinct,
    covar_pop as covar_pop,
    covar_samp as covar_samp,
    cume_dist as cume_dist,
    current_date as current_date,
    current_timestamp as current_timestamp,
    current_user as current_user,
    date_format as date_format,
    date_trunc as date_trunc,
    dateadd as dateadd,
    degrees as degrees,
    dense_rank as dense_rank,
    desc as desc,
    desc_nulls_first as desc_nulls_first,
    desc_nulls_last as desc_nulls_last,
    exp as exp,
    explode as explode,
    expr as expr,
    extract as extract,
    factorial as factorial,
    floor as floor,
    greatest as greatest,
    ifnull as ifnull,
    initcap as initcap,
    input_file_name as input_file_name,
    instr as instr,
    lag as lag,
    lcase as lcase,
    lead as lead,
    least as least,
    length as length,
    levenshtein as levenshtein,
    lit as lit,
    ln as ln,
    locate as locate,
    log as log,
    log10 as log10,
    log2 as log2,
    lower as lower,
    lpad as lpad,
    ltrim as ltrim,
    max as max,
    md5 as md5,
    mean as mean,
    min as min,
    now as now,
    nth_value as nth_value,
    ntile as ntile,
    nullif as nullif,
    nvl as nvl,
    nvl2 as nvl2,
    octet_length as octet_length,
    overlay as overlay,
    percent_rank as percent_rank,
    percentile as percentile,
    pow as pow,
    power as power,
    radians as radians,
    rank as rank,
    regexp_like as regexp_like,
    repeat as repeat,
    reverse as reverse,
    rlike as rlike,
    row_number as row_number,
    rpad as rpad,
    rtrim as rtrim,
    shiftLeft as shiftLeft,
    shiftRight as shiftRight,
    shiftleft as shiftleft,
    shiftright as shiftright,
    sign as sign,
    signum as signum,
    sin as sin,
    sinh as sinh,
    size as size,
    soundex as soundex,
    sqrt as sqrt,
    startswith as startswith,
    stddev as stddev,
    stddev_pop as stddev_pop,
    stddev_samp as stddev_samp,
    substring as substring,
    sum as sum,
    sumDistinct as sumDistinct,
    sum_distinct as sum_distinct,
    tan as tan,
    tanh as tanh,
    timestamp_seconds as timestamp_seconds,
    toDegrees as toDegrees,
    toRadians as toRadians,
    to_date as to_date,
    to_number as to_number,
    to_timestamp as to_timestamp,
    translate as translate,
    trim as trim,
    trunc as trunc,
    ucase as ucase,
    unix_date as unix_date,
    upper as upper,
    user as user,
    var_pop as var_pop,
    var_samp as var_samp,
    variance as variance,
    when as when,
)



import pytest

from sqlframe.standalone import Column as StandaloneColumn
from sqlframe.standalone import Row as StandaloneRow
from sqlframe.standalone import (
    StandaloneCatalog,
    StandaloneDataFrame,
    StandaloneDataFrameNaFunctions,
    StandaloneDataFrameReader,
    StandaloneDataFrameStatFunctions,
    StandaloneDataFrameWriter,
    StandaloneGroupedData,
    StandaloneSession,
    StandaloneUDFRegistration,
)
from sqlframe.standalone import Window as StandaloneWindow
from sqlframe.standalone import WindowSpec as StandaloneWindowSpec
from sqlframe.standalone import functions as StandaloneF
from sqlframe.standalone import types as StandaloneTypes


@pytest.mark.forked
def test_activate_standalone(check_pyspark_imports):
    check_pyspark_imports(
        "standalone",
        sqlf_session=StandaloneSession,
        sqlf_catalog=StandaloneCatalog,
        sqlf_column=StandaloneColumn,
        sqlf_dataframe=StandaloneDataFrame,
        sqlf_grouped_data=StandaloneGroupedData,
        sqlf_window=StandaloneWindow,
        sqlf_window_spec=StandaloneWindowSpec,
        sqlf_functions=StandaloneF,
        sqlf_types=StandaloneTypes,
        sqlf_udf_registration=StandaloneUDFRegistration,
        sqlf_dataframe_reader=StandaloneDataFrameReader,
        sqlf_dataframe_writer=StandaloneDataFrameWriter,
        sqlf_dataframe_na_functions=StandaloneDataFrameNaFunctions,
        sqlf_dataframe_stat_functions=StandaloneDataFrameStatFunctions,
        sqlf_row=StandaloneRow,
    )



# This code is based on code from Apache Spark under the license found in the LICENSE file located in the 'sqlframe' folder.

from __future__ import annotations

import typing as t

from sqlframe.base.operations import Operation, group_operation, operation

if t.TYPE_CHECKING:
    from sqlframe.base.column import Column
    from sqlframe.base.session import DF
else:
    DF = t.TypeVar("DF")


# https://spark.apache.org/docs/latest/sql-ref-syntax-qry-select-groupby.html
# https://stackoverflow.com/questions/37975227/what-is-the-difference-between-cube-rollup-and-groupby-operators
class _BaseGroupedData(t.Generic[DF]):
    def __init__(
        self,
        df: DF,
        group_by_cols: t.Union[t.List[Column], t.List[t.List[Column]]],
        last_op: Operation,
    ):
        self._df = df.copy()
        self.session = df.session
        self.last_op = last_op
        self.group_by_cols = group_by_cols

    def _get_function_applied_columns(
        self, func_name: str, cols: t.Tuple[str, ...]
    ) -> t.List[Column]:
        from sqlframe.base import functions as F

        func_name = func_name.lower()
        return [
            getattr(F, func_name)(name).alias(
                self.session._sanitize_column_name(f"{func_name}({name})")
            )
            for name in cols
        ]

    @group_operation(Operation.SELECT)
    def agg(self, *exprs: t.Union[Column, t.Dict[str, str]]) -> DF:
        from sqlframe.base.column import Column

        columns = (
            [Column(f"{agg_func}({column_name})") for column_name, agg_func in exprs[0].items()]
            if isinstance(exprs[0], dict)
            else exprs
        )
        cols = self._df._ensure_and_normalize_cols(columns)

        if not self.group_by_cols or not isinstance(self.group_by_cols[0], (list, tuple, set)):
            expression = self._df.expression.group_by(
                *[x.expression for x in self.group_by_cols]  # type: ignore
            ).select(*[x.expression for x in self.group_by_cols + cols], append=False)  # type: ignore
            group_by_cols = self.group_by_cols
        else:
            from sqlglot import exp

            expression = self._df.expression
            all_grouping_sets = []
            group_by_cols = []
            for grouping_set in self.group_by_cols:
                all_grouping_sets.append(
                    exp.Tuple(expressions=[x.expression for x in grouping_set])  # type: ignore
                )
                group_by_cols.extend(grouping_set)  # type: ignore
            group_by_cols = list(dict.fromkeys(group_by_cols))
            group_by = exp.Group(grouping_sets=[exp.GroupingSets(expressions=all_grouping_sets)])
            expression.set("group", group_by)
        for col in cols:
            # Spark supports having an empty grouping_id which means all of the columns but other dialects
            # like duckdb don't support this so we expand the grouping_id to include all of the columns
            if col.column_expression.this == "GROUPING_ID":
                col.column_expression.set("expressions", [x.expression for x in group_by_cols])  # type: ignore
        expression = expression.select(*[x.expression for x in group_by_cols + cols], append=False)  # type: ignore
        return self._df.copy(expression=expression)

    def count(self) -> DF:
        from sqlframe.base import functions as F

        return self.agg(F.count("*").alias("count"))

    def mean(self, *cols: str) -> DF:
        return self.avg(*cols)

    def avg(self, *cols: str) -> DF:
        return self.agg(*self._get_function_applied_columns("avg", cols))

    def max(self, *cols: str) -> DF:
        return self.agg(*self._get_function_applied_columns("max", cols))

    def min(self, *cols: str) -> DF:
        return self.agg(*self._get_function_applied_columns("min", cols))

    def sum(self, *cols: str) -> DF:
        return self.agg(*self._get_function_applied_columns("sum", cols))

    def pivot(self, *cols: str) -> DF:
        raise NotImplementedError("Sum distinct is not currently implemented")



import typing as t
from pathlib import Path

import duckdb
import pytest
from pyspark.sql import SparkSession as PySparkSession

from sqlframe.base.session import _BaseSession
from tests.common_fixtures import load_tpcds

pytest_plugins = ["tests.common_fixtures", "tests.integration.fixtures"]


@pytest.mark.parametrize(
    "num",
    list(range(1, 100)),
)
def test_tpcds(
    num: int,
    pyspark_session: PySparkSession,
    duckdb_session: _BaseSession,
    gen_tpcds: t.List[Path],
    compare_frames: t.Callable,
):
    if num in [16, 32, 50, 62, 92, 94, 95, 99]:
        pytest.skip(f"TPCDS{num} is not supported by PySpark due to spaces in column names")
    load_tpcds(gen_tpcds, duckdb_session)
    with open(f"tests/fixtures/tpcds/tpcds{num}.sql") as f:
        query = f.read()
    df = pyspark_session.sql(query)
    dfs = duckdb_session.sql(query)
    compare_frames(df, dfs, no_empty=False)



import pytest

from sqlframe.duckdb import Column as DuckDBColumn
from sqlframe.duckdb import (
    DuckDBCatalog,
    DuckDBDataFrame,
    DuckDBDataFrameNaFunctions,
    DuckDBDataFrameReader,
    DuckDBDataFrameStatFunctions,
    DuckDBDataFrameWriter,
    DuckDBGroupedData,
    DuckDBSession,
    DuckDBUDFRegistration,
)
from sqlframe.duckdb import Row as DuckDBRow
from sqlframe.duckdb import Window as DuckDBWindow
from sqlframe.duckdb import WindowSpec as DuckDBWindowSpec
from sqlframe.duckdb import functions as DuckDBF
from sqlframe.duckdb import types as DuckDBTypes


@pytest.mark.forked
def test_activate_duckdb(check_pyspark_imports):
    check_pyspark_imports(
        "duckdb",
        sqlf_session=DuckDBSession,
        sqlf_catalog=DuckDBCatalog,
        sqlf_column=DuckDBColumn,
        sqlf_dataframe=DuckDBDataFrame,
        sqlf_grouped_data=DuckDBGroupedData,
        sqlf_window=DuckDBWindow,
        sqlf_window_spec=DuckDBWindowSpec,
        sqlf_functions=DuckDBF,
        sqlf_types=DuckDBTypes,
        sqlf_udf_registration=DuckDBUDFRegistration,
        sqlf_dataframe_reader=DuckDBDataFrameReader,
        sqlf_dataframe_writer=DuckDBDataFrameWriter,
        sqlf_dataframe_na_functions=DuckDBDataFrameNaFunctions,
        sqlf_dataframe_stat_functions=DuckDBDataFrameStatFunctions,
        sqlf_row=DuckDBRow,
    )



import inspect
import sys

import sqlframe.base.functions

module = sys.modules["sqlframe.base.functions"]
globals().update(
    {
        name: func
        for name, func in inspect.getmembers(module, inspect.isfunction)
        if hasattr(func, "unsupported_engines")
        and "redshift" not in func.unsupported_engines
        and "*" not in func.unsupported_engines
    }
)


from sqlframe.base.function_alternatives import e_literal as e  # noqa



import datetime

import pytest

from sqlframe.base import types
from sqlframe.snowflake import SnowflakeDataFrame, SnowflakeSession

pytest_plugins = ["tests.integration.fixtures"]
pytestmark = [
    pytest.mark.snowflake,
    pytest.mark.xdist_group("snowflake_tests"),
]


@pytest.fixture()
def snowflake_datatypes(snowflake_session: SnowflakeSession) -> SnowflakeDataFrame:
    return snowflake_session.createDataFrame(
        [
            (
                1,
                2.0,
                "foo",
                {"a": 1},
                [types.Row(a=1, b=2)],
                [1, 2, 3],
                types.Row(a=1),
                datetime.date(2022, 1, 1),
                datetime.datetime(2022, 1, 1, 0, 0, 0),
                datetime.datetime(2022, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc),
                True,
            )
        ],
        [
            "bigint_col",
            "double_col",
            "string_col",
            "map_string_bigint__col",
            "array_struct_a_bigint_b_bigint__",
            "array_bigint__col",
            "struct_a_bigint__col",
            "date_col",
            "timestamp_col",
            "timestamptz_col",
            "boolean_col",
        ],
    )


def test_print_schema_basic(snowflake_employee: SnowflakeDataFrame, capsys):
    snowflake_employee.printSchema()
    captured = capsys.readouterr()
    assert (
        captured.out.strip()
        == """
root
 |-- employee_id: decimal(38, 0) (nullable = true)
 |-- fname: varchar(16777216) (nullable = true)
 |-- lname: varchar(16777216) (nullable = true)
 |-- age: decimal(38, 0) (nullable = true)
 |-- store_id: decimal(38, 0) (nullable = true)""".strip()
    )


def test_print_schema_nested(snowflake_datatypes: SnowflakeDataFrame, capsys):
    snowflake_datatypes.printSchema()
    captured = capsys.readouterr()
    assert (
        captured.out.strip()
        == """
root
 |-- bigint_col: decimal(38, 0) (nullable = true)
 |-- double_col: float (nullable = true)
 |-- string_col: varchar(16777216) (nullable = true)
 |-- map_string_bigint__col: map<varchar(16777216), decimal(38, 0)> (nullable = true)
 |    |-- key: varchar(16777216) (nullable = true)
 |    |-- value: decimal(38, 0) (nullable = true)
 |-- array_struct_a_bigint_b_bigint__: array<object<a decimal(38, 0), b decimal(38, 0)>> (nullable = true)
 |    |-- element: object<a decimal(38, 0), b decimal(38, 0)> (nullable = true)
 |    |    |-- a: decimal(38, 0) (nullable = true)
 |    |    |-- b: decimal(38, 0) (nullable = true)
 |-- array_bigint__col: array<decimal(38, 0)> (nullable = true)
 |    |-- element: decimal(38, 0) (nullable = true)
 |-- struct_a_bigint__col: object<a decimal(38, 0)> (nullable = true)
 |    |-- a: decimal(38, 0) (nullable = true)
 |-- date_col: date (nullable = true)
 |-- timestamp_col: timestamp (nullable = true)
 |-- timestamptz_col: timestamp (nullable = true)
 |-- boolean_col: boolean (nullable = true)""".strip()
    )


def test_schema(snowflake_employee: SnowflakeDataFrame):
    assert isinstance(snowflake_employee.schema, types.StructType)
    struct_fields = list(snowflake_employee.schema)
    assert len(struct_fields) == 5
    assert struct_fields[0].name == "employee_id"
    assert struct_fields[0].dataType == types.DecimalType(38, 0)
    assert struct_fields[1].name == "fname"
    assert struct_fields[1].dataType == types.VarcharType(16777216)
    assert struct_fields[2].name == "lname"
    assert struct_fields[2].dataType == types.VarcharType(16777216)
    assert struct_fields[3].name == "age"
    assert struct_fields[3].dataType == types.DecimalType(38, 0)
    assert struct_fields[4].name == "store_id"
    assert struct_fields[4].dataType == types.DecimalType(38, 0)


def test_schema_nested(snowflake_datatypes: SnowflakeDataFrame):
    assert isinstance(snowflake_datatypes.schema, types.StructType)
    struct_fields = list(snowflake_datatypes.schema)
    assert len(struct_fields) == 11
    assert struct_fields[0].name == "bigint_col"
    assert struct_fields[0].dataType == types.DecimalType(38, 0)
    assert struct_fields[1].name == "double_col"
    assert struct_fields[1].dataType == types.FloatType()
    assert struct_fields[2].name == "string_col"
    assert struct_fields[2].dataType == types.VarcharType(16777216)
    assert struct_fields[3].name == "map_string_bigint__col"
    assert struct_fields[3].dataType == types.MapType(
        types.VarcharType(16777216),
        types.DecimalType(38, 0),
    )
    assert struct_fields[4].name == "array_struct_a_bigint_b_bigint__"
    assert struct_fields[4].dataType == types.ArrayType(
        types.StructType(
            [
                types.StructField(
                    "a",
                    types.DecimalType(38, 0),
                ),
                types.StructField(
                    "b",
                    types.DecimalType(38, 0),
                ),
            ]
        ),
    )
    assert struct_fields[5].name == "array_bigint__col"
    assert struct_fields[5].dataType == types.ArrayType(
        types.DecimalType(38, 0),
    )
    assert struct_fields[6].name == "struct_a_bigint__col"
    assert struct_fields[6].dataType == types.StructType(
        [
            types.StructField(
                "a",
                types.DecimalType(38, 0),
            ),
        ]
    )
    assert struct_fields[7].name == "date_col"
    assert struct_fields[7].dataType == types.DateType()
    assert struct_fields[8].name == "timestamp_col"
    assert struct_fields[8].dataType == types.TimestampType()
    assert struct_fields[9].name == "timestamptz_col"
    assert struct_fields[9].dataType == types.TimestampType()
    assert struct_fields[10].name == "boolean_col"
    assert struct_fields[10].dataType == types.BooleanType()



import sys
from unittest.mock import MagicMock

import findspark
import pytest

from sqlframe import activate, deactivate
from sqlframe import testing as SQLFrameTesting


@pytest.mark.forked
def test_activate_testing():
    activate()
    findspark.init()
    from pyspark import testing

    assert testing == SQLFrameTesting
    assert testing.assertDataFrameEqual == SQLFrameTesting.assertDataFrameEqual
    assert testing.assertSchemaEqual == SQLFrameTesting.assertSchemaEqual
    from pyspark.testing import assertDataFrameEqual, assertSchemaEqual

    assert assertDataFrameEqual == SQLFrameTesting.assertDataFrameEqual
    assert assertSchemaEqual == SQLFrameTesting.assertSchemaEqual
    import pyspark.testing as testing

    assert testing == SQLFrameTesting
    assert testing.assertDataFrameEqual == SQLFrameTesting.assertDataFrameEqual


@pytest.mark.forked
def test_activate_no_engine():
    activate()
    findspark.init()
    # A way that people check if pyspark is available
    from pyspark import context

    assert isinstance(context, MagicMock)

    deactivate()

    from pyspark import context

    assert not isinstance(context, MagicMock)
    assert context is not MagicMock
    assert context is not None



from __future__ import annotations

import logging
import sys
import typing as t

from sqlframe.base.catalog import Column as CatalogColumn
from sqlframe.base.dataframe import (
    _BaseDataFrame,
    _BaseDataFrameNaFunctions,
    _BaseDataFrameStatFunctions,
)
from sqlframe.base.mixins.dataframe_mixins import NoCachePersistSupportMixin
from sqlframe.base.util import normalize_string
from sqlframe.snowflake.group import SnowflakeGroupedData

if t.TYPE_CHECKING:
    from sqlframe.snowflake.readwriter import SnowflakeDataFrameWriter
    from sqlframe.snowflake.session import SnowflakeSession


logger = logging.getLogger(__name__)


class SnowflakeDataFrameNaFunctions(_BaseDataFrameNaFunctions["SnowflakeDataFrame"]):
    pass


class SnowflakeDataFrameStatFunctions(_BaseDataFrameStatFunctions["SnowflakeDataFrame"]):
    pass


class SnowflakeDataFrame(
    NoCachePersistSupportMixin,
    _BaseDataFrame[
        "SnowflakeSession",
        "SnowflakeDataFrameWriter",
        "SnowflakeDataFrameNaFunctions",
        "SnowflakeDataFrameStatFunctions",
        "SnowflakeGroupedData",
    ],
):
    _na = SnowflakeDataFrameNaFunctions
    _stat = SnowflakeDataFrameStatFunctions
    _group_data = SnowflakeGroupedData

    @property
    def _typed_columns(self) -> t.List[CatalogColumn]:
        df = self._convert_leaf_to_cte()
        df = df.limit(0)
        df.collect()
        query_id = self.session._cur.sfqid
        columns = []
        for row in self.session._collect(f"DESCRIBE RESULT '{query_id}'"):
            null_row = [x for x in row.__fields__ if "?" in x][0]
            columns.append(
                CatalogColumn(
                    name=normalize_string(row.name, from_dialect="execution", to_dialect="output"),
                    dataType=normalize_string(
                        row.type, from_dialect="execution", to_dialect="output", is_datatype=True
                    ),
                    nullable=row[null_row] == "Y",
                    description=row.comment,
                    isPartition=False,
                    isBucket=False,
                )
            )
        return columns



from __future__ import annotations

import os
import typing as t

import pytest
from sqlglot import exp

from sqlframe.databricks.session import DatabricksSession

if t.TYPE_CHECKING:
    from databricks.sql import Connection as DatabricksConnection

pytest_plugins = ["tests.common_fixtures"]
pytestmark = [
    pytest.mark.databricks,
    pytest.mark.xdist_group("databricks_tests"),
]


@pytest.fixture
def cleanup_connector() -> t.Iterator[DatabricksConnection]:
    from databricks.sql import connect

    conn = connect(
        server_hostname=os.environ["SQLFRAME_DATABRICKS_SERVER_HOSTNAME"],
        http_path=os.environ["SQLFRAME_DATABRICKS_HTTP_PATH"],
        access_token=os.environ["SQLFRAME_DATABRICKS_ACCESS_TOKEN"],
        auth_type="access_token",
        catalog=os.environ["SQLFRAME_DATABRICKS_CATALOG"],
        schema=os.environ["SQLFRAME_DATABRICKS_SCHEMA"],
    )
    conn.cursor().execute("CREATE SCHEMA IF NOT EXISTS db1")
    conn.cursor().execute("DROP TABLE IF EXISTS db1.test_table")
    conn.cursor().execute("CREATE TABLE IF NOT EXISTS db1.test_table (cola INT, colb STRING)")

    yield conn
    conn.cursor().execute("DROP TABLE IF EXISTS db1.test_table")


def test_session_from_config(cleanup_connector: DatabricksConnection):
    session = DatabricksSession.builder.config("sqlframe.conn", cleanup_connector).getOrCreate()
    columns = session.catalog.get_columns("sqlframe.db1.test_table")
    assert columns == {
        "cola": exp.DataType.build("INT", dialect=session.output_dialect),
        "colb": exp.DataType.build("STRING", dialect=session.output_dialect),
    }



from __future__ import annotations

import typing as t

from sqlframe.base.session import _BaseSession
from sqlframe.bigquery.catalog import BigQueryCatalog
from sqlframe.bigquery.dataframe import BigQueryDataFrame
from sqlframe.bigquery.readwriter import (
    BigQueryDataFrameReader,
    BigQueryDataFrameWriter,
)
from sqlframe.bigquery.udf import BigQueryUDFRegistration

if t.TYPE_CHECKING:
    from google.cloud.bigquery.client import Client as BigQueryClient
    from google.cloud.bigquery.dbapi.connection import Connection as BigQueryConnection
else:
    BigQueryClient = t.Any
    BigQueryConnection = t.Any


class BigQuerySession(
    _BaseSession[  # type: ignore
        BigQueryCatalog,
        BigQueryDataFrameReader,
        BigQueryDataFrameWriter,
        BigQueryDataFrame,
        BigQueryConnection,
        BigQueryUDFRegistration,
    ],
):
    _catalog = BigQueryCatalog
    _reader = BigQueryDataFrameReader
    _writer = BigQueryDataFrameWriter
    _df = BigQueryDataFrame
    _udf_registration = BigQueryUDFRegistration

    QUALIFY_INFO_SCHEMA_WITH_DATABASE = True
    SANITIZE_COLUMN_NAMES = True

    def __init__(
        self, conn: t.Optional[BigQueryConnection] = None, default_dataset: t.Optional[str] = None
    ):
        from google.cloud import bigquery
        from google.cloud.bigquery.dbapi import connect

        if not hasattr(self, "_conn"):
            super().__init__(conn or connect())
            if self._client.default_query_job_config is None:
                self._client.default_query_job_config = bigquery.QueryJobConfig()
            self.default_dataset = default_dataset

    @property
    def _client(self) -> BigQueryClient:
        assert self._connection
        return self._connection._client

    @property
    def default_dataset(self) -> t.Optional[str]:
        return self._default_dataset

    @default_dataset.setter
    def default_dataset(self, dataset: str) -> None:
        self._default_dataset = dataset
        self._client.default_query_job_config.default_dataset = dataset  # type: ignore

    @property
    def default_project(self) -> str:
        return self._client.project

    @default_project.setter
    def default_project(self, project: str) -> None:
        self._client.project = project

    @classmethod
    def _try_get_map(cls, value: t.Any) -> t.Optional[t.Dict[str, t.Any]]:
        return None

    class Builder(_BaseSession.Builder):
        DEFAULT_EXECUTION_DIALECT = "bigquery"

        @property
        def session(self) -> BigQuerySession:
            return BigQuerySession(**self._session_kwargs)

        def getOrCreate(self) -> BigQuerySession:
            return super().getOrCreate()  # type: ignore

    builder = Builder()
